from __future__ import annotations

from typing import TYPE_CHECKING

import discord

from pylav.helpers.misc import TimedFeature
from pylav.logging import getLogger
from pylav.storage.models.player.config import PlayerConfig

if TYPE_CHECKING:
    from pylav.core.client import Client
LOGGER = getLogger("PyLav.Database.Controller.Player.Config")


class PlayerConfigController:
    __slots__ = ("_client",)

    def __init__(self, client: Client) -> None:
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    async def initialize_global_config(self) -> None:
        await PlayerConfig.create_global(bot=self.client.bot.user.id)

    def get_global_config(self) -> PlayerConfig:
        return PlayerConfig(bot=self.client.bot.user.id, id=0)

    def get_config(self, guild_id: int) -> PlayerConfig:
        return PlayerConfig(bot=self.client.bot.user.id, id=guild_id)

    async def reset_to_default(self, guild_id: int) -> None:
        await self.get_config(guild_id=guild_id).delete()

    async def get_volume(self, guild_id: int) -> int:
        global_vol = await self.get_global_config().fetch_volume()
        server_vol = await self.get_config(guild_id=guild_id).fetch_volume()
        return global_vol if global_vol < server_vol else server_vol

    async def get_max_volume(self, guild_id: int) -> int:
        global_vol = await self.get_global_config().fetch_max_volume()
        server_vol = await self.get_config(guild_id=guild_id).fetch_max_volume()
        return global_vol if global_vol < server_vol else server_vol

    async def get_shuffle(self, guild_id: int) -> bool:
        if await self.get_global_config().fetch_shuffle() is False:
            return False
        return await self.get_config(guild_id=guild_id).fetch_shuffle()

    async def get_auto_shuffle(self, guild_id: int) -> bool:
        if await self.get_global_config().fetch_auto_shuffle() is False:
            return False
        return await self.get_config(guild_id=guild_id).fetch_auto_shuffle()

    async def get_self_deaf(self, guild_id: int) -> bool:
        if await self.get_global_config().fetch_self_deaf() is True:
            return True
        return await self.get_config(guild_id=guild_id).fetch_self_deaf()

    async def get_empty_queue_dc(self, guild_id: int) -> TimedFeature:
        if (global_empty_queue_dc := await self.get_global_config().fetch_empty_queue_dc()).enabled is True:
            return global_empty_queue_dc
        return await self.get_config(guild_id=guild_id).fetch_empty_queue_dc()

    async def get_alone_dc(self, guild_id: int) -> TimedFeature:
        if (global_alone_dc := await self.get_global_config().fetch_alone_dc()).enabled is True:
            return global_alone_dc
        return await self.get_config(guild_id=guild_id).fetch_alone_dc()

    async def get_alone_pause(self, guild_id: int) -> TimedFeature:
        if (global_alone_pause := await self.get_global_config().fetch_alone_pause()).enabled is True:
            return global_alone_pause
        return await self.get_config(guild_id=guild_id).fetch_alone_pause()

    async def get_auto_play(self, guild_id: int) -> bool:
        if await self.get_global_config().fetch_auto_play() is False:
            return False
        return await self.get_config(guild_id=guild_id).fetch_auto_play()

    async def is_dj(
        self,
        user: discord.Member,
        guild: discord.Guild,
        *,
        additional_role_ids: list = None,
        additional_user_ids: list = None,
    ) -> bool:
        if additional_user_ids and user.id in additional_user_ids:
            return True
        if additional_role_ids and any(r.id in additional_role_ids for r in user.roles):
            return True
        return await self.get_config(guild_id=guild.id).is_dj(
            user=user, additional_role_ids=None, additional_user_ids=None
        )



import dataclasses

from pylav.nodes.api.responses.filters import plugins
from pylav.nodes.api.responses.filters.channel_mix import ChannelMix as ChannelMix
from pylav.nodes.api.responses.filters.distortion import Distortion as Distortion
from pylav.nodes.api.responses.filters.equalizer import Equalizer as Equalizer
from pylav.nodes.api.responses.filters.equalizer import EqualizerBand as EqualizerBand
from pylav.nodes.api.responses.filters.karaoke import Karaoke as Karaoke
from pylav.nodes.api.responses.filters.low_pass import LowPass as LowPass
from pylav.nodes.api.responses.filters.plugins import *
from pylav.nodes.api.responses.filters.rotation import Rotation as Rotation
from pylav.nodes.api.responses.filters.timescale import Timescale as Timescale
from pylav.nodes.api.responses.filters.tremolo import Tremolo as Tremolo
from pylav.nodes.api.responses.filters.vibrato import Vibrato as Vibrato

__all__ = (
    "Filters",
    "ChannelMix",
    "Distortion",
    "Echo",
    "Equalizer",
    "EqualizerBand",
    "Karaoke",
    "LowPass",
    "Rotation",
    "Timescale",
    "Tremolo",
    "Vibrato",
    *plugins.__all__,
)

from pylav.type_hints.dict_typing import JSON_DICT_TYPE
from pylav.type_hints.generics import ANY_GENERIC_TYPE


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Filters:
    volume: float | None = None
    equalizer: Equalizer | None = dataclasses.field(default_factory=list)
    karaoke: Karaoke | None = dataclasses.field(default_factory=dict)
    timescale: Timescale | None = dataclasses.field(default_factory=dict)
    tremolo: Tremolo | None = dataclasses.field(default_factory=dict)
    vibrato: Vibrato | None = dataclasses.field(default_factory=dict)
    rotation: Rotation | None = dataclasses.field(default_factory=dict)
    distortion: Distortion | None = dataclasses.field(default_factory=dict)
    channelMix: ChannelMix | None = dataclasses.field(default_factory=dict)
    lowPass: LowPass | None = dataclasses.field(default_factory=dict)
    echo: Echo | None = dataclasses.field(default_factory=dict)

    def to_dict(self) -> JSON_DICT_TYPE:
        response: JSON_DICT_TYPE = {"volume": self.volume}
        for filter_name in [
            "equalizer",
            "karaoke",
            "timescale",
            "tremolo",
            "vibrato",
            "rotation",
            "distortion",
            "channelMix",
            "lowPass",
            "echo",
        ]:
            response = self._process_filter(filter_name, response)
        return response

    def _process_filter(self, name: str, response: ANY_GENERIC_TYPE) -> ANY_GENERIC_TYPE:
        match name:
            case "equalizer":
                return self._process_equalizer(response)
            case "karaoke":
                return self._process_filter_object(name, self.karaoke, Karaoke, response)
            case "timescale":
                return self._process_filter_object(name, self.timescale, Timescale, response)
            case "tremolo":
                return self._process_filter_object(name, self.tremolo, Tremolo, response)
            case "vibrato":
                return self._process_filter_object(name, self.vibrato, Vibrato, response)
            case "rotation":
                return self._process_filter_object(name, self.rotation, Rotation, response)
            case "distortion":
                return self._process_filter_object(name, self.distortion, Distortion, response)
            case "channelMix":
                return self._process_filter_object(name, self.channelMix, ChannelMix, response)
            case "lowPass":
                return self._process_filter_object(name, self.lowPass, LowPass, response)
            case "echo":
                return self._process_filter_object(name, self.echo, Echo, response)
            case __:
                return response

    def _process_equalizer(self, response: ANY_GENERIC_TYPE) -> ANY_GENERIC_TYPE:
        if isinstance(self.equalizer, list):
            response["equalizer"] = [
                band.to_dict() if isinstance(band, EqualizerBand) else band for band in self.equalizer
            ]
        else:
            response["equalizer"] = None
        return response

    @staticmethod
    def _process_filter_object(
        name: str,
        filter_name: Karaoke
        | Timescale
        | Tremolo
        | Vibrato
        | Rotation
        | Distortion
        | ChannelMix
        | LowPass
        | Echo
        | None,
        cls: type[Karaoke | Timescale | Tremolo | Vibrato | Rotation | Distortion | ChannelMix | LowPass | Echo],
        response: ANY_GENERIC_TYPE,
    ) -> ANY_GENERIC_TYPE:
        if isinstance(filter_name, cls):
            response[name] = filter_name.to_dict()
        else:
            response[name] = None
        return response



from __future__ import annotations

import os
import pathlib

import yaml

from pylav.logging import getLogger

if full_path := os.getenv("PYLAV__YAML_CONFIG"):
    ENV_FILE = pathlib.Path(full_path)
else:
    ENV_FILE = pathlib.Path.home() / "pylav.yaml"

LOGGER = getLogger("PyLav.Environment")


def build_from_envvars() -> None:
    from pylav.constants.config.env_var import (
        DATA_FOLDER,
        DEFAULT_SEARCH_SOURCE,
        EXTERNAL_UNMANAGED_HOST,
        EXTERNAL_UNMANAGED_NAME,
        EXTERNAL_UNMANAGED_PASSWORD,
        EXTERNAL_UNMANAGED_PORT,
        EXTERNAL_UNMANAGED_SSL,
        FALLBACK_POSTGREST_HOST,
        JAVA_EXECUTABLE,
        LINKED_BOT_IDS,
        LOCAL_TRACKS_FOLDER,
        MANAGED_NODE_APPLE_MUSIC_API_KEY,
        MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE,
        MANAGED_NODE_DEEZER_KEY,
        MANAGED_NODE_SPOTIFY_CLIENT_ID,
        MANAGED_NODE_SPOTIFY_CLIENT_SECRET,
        MANAGED_NODE_SPOTIFY_COUNTRY_CODE,
        MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN,
        POSTGRES_DATABASE,
        POSTGRES_PASSWORD,
        POSTGRES_PORT,
        POSTGRES_SOCKET,
        POSTGRES_USER,
        READ_CACHING_ENABLED,
        REDIS_FULL_ADDRESS_RESPONSE_CACHE,
        TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
        TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
        TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
        USE_BUNDLED_EXTERNAL_PYLAV_NODE,
    )

    # noinspection SpellCheckingInspection
    data = {
        "PYLAV__POSTGRES_PORT": POSTGRES_PORT,
        "PYLAV__POSTGRES_PASSWORD": POSTGRES_PASSWORD,
        "PYLAV__POSTGRES_USER": POSTGRES_USER,
        "PYLAV__POSTGRES_DB": POSTGRES_DATABASE,
        "PYLAV__POSTGRES_SOCKET": POSTGRES_SOCKET,
        "PYLAV__POSTGRES_HOST": FALLBACK_POSTGREST_HOST,
        "PYLAV__REDIS_FULLADDRESS_RESPONSE_CACHE": REDIS_FULL_ADDRESS_RESPONSE_CACHE,
        "PYLAV__JAVA_EXECUTABLE": JAVA_EXECUTABLE,
        "PYLAV__LINKED_BOT_IDS": LINKED_BOT_IDS,
        "PYLAV__USE_BUNDLED_EXTERNAL_PYLAV_NODE": USE_BUNDLED_EXTERNAL_PYLAV_NODE,
        "PYLAV__EXTERNAL_UNMANAGED_HOST": EXTERNAL_UNMANAGED_HOST,
        "PYLAV__EXTERNAL_UNMANAGED_PORT": EXTERNAL_UNMANAGED_PORT,
        "PYLAV__EXTERNAL_UNMANAGED_PASSWORD": EXTERNAL_UNMANAGED_PASSWORD,
        "PYLAV__EXTERNAL_UNMANAGED_SSL": EXTERNAL_UNMANAGED_SSL,
        "PYLAV__EXTERNAL_UNMANAGED_NAME": EXTERNAL_UNMANAGED_NAME,
        "PYLAV__TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS": TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
        "PYLAV__TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS": TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
        "PYLAV__TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS": TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
        "PYLAV__READ_CACHING_ENABLED": READ_CACHING_ENABLED,
        "PYLAV__DEFAULT_SEARCH_SOURCE": DEFAULT_SEARCH_SOURCE,
        "PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_ID": MANAGED_NODE_SPOTIFY_CLIENT_ID,
        "PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_SECRET": MANAGED_NODE_SPOTIFY_CLIENT_SECRET,
        "PYLAV__MANAGED_NODE_SPOTIFY_COUNTRY_CODE": MANAGED_NODE_SPOTIFY_COUNTRY_CODE,
        "PYLAV__MANAGED_NODE_APPLE_MUSIC_API_KEY": MANAGED_NODE_APPLE_MUSIC_API_KEY,
        "PYLAV__MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE": MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE,
        "PYLAV__MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN": MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN,
        "PYLAV__MANAGED_NODE_DEEZER_KEY": MANAGED_NODE_DEEZER_KEY,
        "PYLAV__LOCAL_TRACKS_FOLDER": LOCAL_TRACKS_FOLDER,
        "PYLAV__DATA_FOLDER": DATA_FOLDER,
    }
    with ENV_FILE.open(mode="w") as file:
        LOGGER.debug("Creating %s with the following content: %r", ENV_FILE, data)
        yaml.safe_dump(data, file, default_flow_style=False, sort_keys=False, encoding="utf-8")



from __future__ import annotations

from pylav.players.filters.misc import FilterMixin


class Echo(FilterMixin):
    __slots__ = ("_delay", "_decay", "_default")

    def __init__(self, delay: float | None = None, decay: float | None = None) -> None:
        super().__init__()
        self.delay = delay
        self.decay = decay

    def to_dict(self) -> dict[str, float | bool | None]:
        return {
            "delay": self.delay,
            "decay": self.decay,
        }

    @classmethod
    def from_dict(cls, data: dict[str, float | bool | None]) -> Echo:
        return cls(delay=data["delay"], decay=data["decay"])

    def __repr__(self) -> str:
        return f"<Echo: delay={self.delay}, decay={self.decay}>"

    @property
    def delay(self) -> float | None:
        return self._delay

    @delay.setter
    def delay(self, v: float | None) -> None:
        if v is None:
            self._delay = None
            return
        if v < 0:
            raise ValueError(f"Delay must be must be greater than 0, not {v}")
        self._delay = v

    @property
    def decay(self) -> float | None:
        return self._decay

    @decay.setter
    def decay(self, v: float | None) -> None:
        if v is None:
            self._decay = v
            return
        if not (0.0 < v <= 1.0):
            raise ValueError(f"Decay must be must be 0.0 < x ≤ 1.0, not {v}")
        self._decay = v

    @classmethod
    def default(cls) -> Echo:
        return cls()

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        response = {}
        if self.delay is not None:
            response["delay"] = self.delay
        if self.decay is not None:
            response["decay"] = self.decay
        return response

    def reset(self) -> None:
        self.delay = self.decay = None



from __future__ import annotations

from pylav.constants.regex import SOURCE_INPUT_MATCH_DEEZER, TIMESTAMP_YOUTUBE, YOUTUBE_TRACK_INDEX


def process_youtube(cls: type[Query], query: str, music: bool) -> Query:
    index = 0
    if match := TIMESTAMP_YOUTUBE.search(query):
        start_time = int(match.group(1))
    else:
        start_time = 0
    _has_index = "&index=" in query
    if _has_index and (match := YOUTUBE_TRACK_INDEX.search(query)):
        index = int(match.group(1)) - 1
    if all(k in query for k in ["&list=", "watch?"]):
        query_type = "playlist"
        index = 0
    elif all(x in query for x in ["playlist?"]):
        query_type = "playlist"
    elif any(k in query for k in ["list="]):
        index = 0
        query_type = "single" if _has_index else "playlist"
    else:
        query_type = "single"
    return cls(
        query,
        "YouTube Music" if music else "YouTube",
        start_time=start_time,
        query_type=query_type,
        index=index,
    )


def process_deezer(cls: type[Query], query: str) -> Query:
    """Process a Deezer query."

    Parameters
    ----------
    cls : QueryT
        The class to instantiate.
    query : str
        The query to process.
    Returns
    -------
    Query
        The processed query.
    """
    search = SOURCE_INPUT_MATCH_DEEZER.search(query)
    if search is None:
        raise ValueError("Invalid Deezer query")
    data = search.groupdict()
    # noinspection SpellCheckingInspection
    query_type = data.get("dztype")
    return cls(
        query,
        "Deezer",
        query_type="single" if query_type == "track" else "album" if query_type == "album" else "playlist",
    )


def process_spotify(cls: type[Query], query: str) -> Query:
    query_type = "single"
    if "/playlist/" in query:
        query_type = "playlist"
    elif "/album/" in query:
        query_type = "album"
    return cls(query, "Spotify", query_type=query_type)


def process_soundcloud(cls: type[Query], query: str) -> Query:
    if "/sets/" in query and "?in=" in query or "/sets/" not in query:
        query_type = "single"
    else:
        query_type = "playlist"
    return cls(query, "SoundCloud", query_type=query_type)


def process_bandcamp(cls: type[Query], query: str) -> Query:
    query_type = "album" if "/album/" in query else "single"
    return cls(query, "Bandcamp", query_type=query_type)


def process_yandex_music(cls: type[Query], query: str) -> Query:
    query_type = "single"
    if "/album/" in query and "/track/" not in query:
        query_type = "album"
    elif "/playlist/" in query:
        query_type = "playlist"
    return cls(query, "Yandex Music", query_type=query_type)


from pylav.players.query.obj import Query  # noqa: E305



from __future__ import annotations

import asyncio
import os

import aiopath
from cashews import Cache

from pylav.extension.m3u.http_client import DefaultHTTPClient, parsed_url
from pylav.extension.m3u.models import M3U8
from pylav.extension.m3u.parser import is_url

# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.


CACHE = Cache("M3U3CACHE")
CACHE.setup("mem://?check_interval=10", size=1_000_000, enable=True)


@CACHE(ttl=600, prefix="m3u_loads")
async def loads(self, content: str, uri: str = None, custom_tags_parser=None):  # noqa
    """
    Given a string with a m3u8 content, returns a M3U8 object.
    Optionally parses a uri to set a correct base_uri on the M3U8 object.
    Raises ValueError if invalid content
    """

    if uri is None:
        return await asyncio.to_thread(M3U8, content, custom_tags_parser=custom_tags_parser)
    else:
        parsed_url(uri)


@CACHE(ttl=600, prefix="m3u_load")
async def load(
    self,  # noqa
    uri: str,
    timeout: float = None,
    headers: dict[str, str] = None,
    custom_tags_parser=None,
    http_client: DefaultHTTPClient = DefaultHTTPClient(),
    verify_ssl: bool = True,
):
    """
    Retrieves the content from a given URI and returns a M3U8 object.
    Raises ValueError if invalid content or IOError if request fails.
    """
    if headers is None:
        headers = {}
    if is_url(uri):
        content, base_uri = await http_client.download(uri=uri, timeout=timeout, headers=headers, verify_ssl=verify_ssl)
        return await asyncio.to_thread(M3U8, content, base_uri=base_uri, custom_tags_parser=custom_tags_parser)
    else:
        return await load_from_file(uri, custom_tags_parser)


async def load_from_file(uri: str, custom_tags_parser=None):
    """Loads a m3u8 file from the filesystem

    Parameters:
    -----------
    uri: str
        uri of the m3u8 file
    custom_tags_parser: callable
        custom tags parser function
    """

    path = aiopath.AsyncPath(uri)
    async with path.open("r", encoding="utf8") as file:
        raw_content = (await file.read()).strip()
    base_uri = os.path.dirname(uri)
    return M3U8(raw_content, base_uri=base_uri, custom_tags_parser=custom_tags_parser)



from __future__ import annotations

import itertools
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import discord
from discord import Emoji, PartialEmoji
from redbot.core.i18n import Translator

from pylav.extension.red.utils import rgetattr
from pylav.helpers import emojis
from pylav.helpers.format.strings import shorten_string
from pylav.storage.models.playlist import Playlist
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

if TYPE_CHECKING:
    from pylav.extension.red.ui.menus.playlist import PlaylistCreationFlow, PlaylistManageFlow
_ = Translator("PyLav", Path(__file__))


class PlaylistDeleteButton(discord.ui.Button):
    view: PlaylistManageFlow

    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.TRASH,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )
        self.view.cancelled = False
        self.view.delete = not self.view.delete
        if self.view.delete:
            response = _("I will delete this playlist permanently once you select “done”.")
        else:
            response = _("I will no longer delete this playlist permanently once you select “done”.")

        await context.send(
            embed=await self.cog.pylav.construct_embed(messageable=interaction, description=response),
            ephemeral=True,
        )


class PlaylistClearButton(discord.ui.Button):
    view: PlaylistManageFlow

    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.CLEAR,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )
        self.view.cancelled = False
        self.view.clear = not self.view.clear
        if self.view.clear:
            response = _("I will remove all tracks from this playlist.")
        else:
            response = _("I will no longer remove all tracks from this playlist.")

        await context.send(
            embed=await self.cog.pylav.construct_embed(messageable=interaction, description=response),
            ephemeral=True,
        )


class PlaylistDownloadButton(discord.ui.Button):
    view: PlaylistManageFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        emoji: str | Emoji | PartialEmoji,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emoji,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )
        async with self.view.playlist.to_yaml(guild=interaction.guild) as (yaml_file, compressed):
            yaml_file: BytesIO
            if compressed == "gzip":
                message = _(
                    "Here is your playlist: {playlist_name_variable_do_not_translate}. "
                    "This file is compressed using Gzip to make it possible to be sent via Discord. "
                    "You can use {url_name_variable_do_not_translate} to decompress it."
                ).format(
                    playlist_name_variable_do_not_translate=await self.view.playlist.get_name_formatted(with_url=True),
                    url_name_variable_do_not_translate="<https://gzip.swimburger.net/>",
                )
            elif compressed == "brotli":
                message = _(
                    "Here is your playlist: {playlist_name_name_variable_do_not_translate}. "
                    "This file is compressed using Brotli to make it possible to be sent via Discord."
                ).format(
                    playlist_name_name_variable_do_not_translate=await self.view.playlist.get_name_formatted(
                        with_url=True
                    )
                )
            else:
                message = _("Here is your playlist: {playlist_name_name_variable_do_not_translate}.").format(
                    playlist_name_name_variable_do_not_translate=await self.view.playlist.get_name_formatted(
                        with_url=True
                    )
                )

            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction,
                    description=message,
                ),
                file=discord.File(
                    filename=f"{await self.view.playlist.fetch_name()}{'.gz' if compressed=='gzip' else '.br' if compressed=='brotli' else ''}.pylav",
                    fp=yaml_file,
                ),
                ephemeral=True,
            )


class PlaylistUpdateButton(discord.ui.Button):
    view: PlaylistManageFlow

    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.UPDATE,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )
        self.view.cancelled = False
        self.view.update = not self.view.update
        if (await self.view.playlist.fetch_url() or self.view.url) and self.view.update:
            response = _("I will update this playlist with the latest tracks, press done to continue.")
        else:
            self.view.update = False
            response = _("I will no longer update this playlist with the latest tracks.")
        await context.send(
            embed=await self.cog.pylav.construct_embed(messageable=interaction, description=response),
            ephemeral=True,
        )


class PlaylistInfoButton(discord.ui.Button):
    view: PlaylistManageFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        emoji: str | Emoji | PartialEmoji,
        playlist: Playlist,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emoji,
            row=row,
        )
        self.cog = cog
        self.playlist = playlist

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )

        from pylav.extension.red.ui.menus.generic import PaginatingMenu
        from pylav.extension.red.ui.sources.playlist import TrackMappingSource

        tracks = await self.view.playlist.fetch_tracks()
        await PaginatingMenu(
            bot=self.cog.bot,
            cog=self.cog,
            source=TrackMappingSource(
                guild_id=interaction.guild.id,
                cog=self.cog,
                author=interaction.user,
                entries=tracks,
                playlist=self.playlist,
            ),
            delete_after_timeout=True,
            starting_page=0,
            original_author=interaction.user,
        ).start(context)


class PlaylistQueueButton(discord.ui.Button):
    view: PlaylistManageFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        emoji: str | Emoji | PartialEmoji,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emoji,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )
        self.view.queue = not self.view.queue
        if self.view.queue:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("I will add the current queue this playlist.")
                ),
                ephemeral=True,
            )
        else:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("I will no longer add the current queue to this playlist.")
                ),
                ephemeral=True,
            )


class PlaylistDedupeButton(discord.ui.Button):
    view: PlaylistManageFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        emoji: str | Emoji | PartialEmoji,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emoji,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)

        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )
        self.view.dedupe = not self.view.dedupe
        if self.view.dedupe:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("I will remove all duplicate tracks from the queue.")
                ),
                ephemeral=True,
            )
        else:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("I will no longer remove duplicate tracks from the queue.")
                ),
                ephemeral=True,
            )


class PlaylistUpsertButton(discord.ui.Button):
    view: PlaylistCreationFlow | PlaylistManageFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        op: Literal["url", "name", "scope", "add", "remove"],
        label: str = None,
        emoji: str | Emoji | PartialEmoji = None,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emoji,
            label=label,
            row=row,
        )
        self.cog = cog
        self.op = op

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if self.view.author.id != interaction.user.id:
            return await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )
        self.view.cancelled = False
        if self.op == "url":
            await self.view.prompt_url(interaction)
        elif self.op == "name":
            await self.view.prompt_name(interaction)
        elif self.op == "scope":
            await self.view.prompt_scope(interaction)
        elif self.op == "add":
            await self.view.prompt_add_tracks(interaction)
        elif self.op == "remove":
            await self.view.prompt_remove_tracks(interaction)


class EnqueuePlaylistButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        row: int = None,
        emoji: Emoji | PartialEmoji = emojis.PLAYLIST,
        playlist: Playlist = None,
    ):
        self.cog = cog
        super().__init__(
            style=style,
            emoji=emoji,
            row=row,
        )
        self.playlist = playlist

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if not self.playlist:
            playlists = await self.cog.pylav.playlist_db_manager.get_all_for_user(
                requester=context.author.id,
                vc=rgetattr(context.author, "voice.channel", None),
                guild=context.guild,
                channel=context.channel,
            )
            playlists = list(itertools.chain.from_iterable(playlists))

            from pylav.extension.red.ui.menus.playlist import PlaylistPickerMenu
            from pylav.extension.red.ui.selectors.playlist import PlaylistPlaySelector
            from pylav.extension.red.ui.sources.playlist import PlaylistPickerSource

            await PlaylistPickerMenu(
                cog=self.cog,
                bot=self.cog.bot,
                selector_cls=PlaylistPlaySelector,
                source=PlaylistPickerSource(
                    guild_id=context.guild.id,
                    cog=self.cog,
                    pages=playlists,
                    message_str=_("Playlists you can currently play."),
                ),
                delete_after_timeout=True,
                clear_buttons_after=True,
                starting_page=0,
                original_author=context.author,
                selector_text=shorten_string(max_length=100, string=_("Pick a playlist.")),
            ).start(context)
        else:
            await self.cog.command_playlist_play.callback(self.cog, interaction, playlist=[self.playlist])
        if hasattr(self.view, "prepare"):
            await self.view.prepare()
            kwargs = await self.view.get_page(self.view.current_page)
            await self.view.message.edit(view=self.view, **kwargs)


class SaveQueuePlaylistButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.QUEUE,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        from pylav.extension.red.ui.modals.playlist import PlaylistSaveModal

        modal = PlaylistSaveModal(self.cog, self, _("What should the playlist name be?"))
        await interaction.response.send_modal(modal)



from __future__ import annotations

import heapq
import pathlib
import re
from collections.abc import Awaitable, Callable
from datetime import timedelta
from functools import wraps
from operator import attrgetter
from typing import TYPE_CHECKING, Any

from cashews import Cache
from discord.app_commands import Choice
from rapidfuzz import fuzz

from pylav.constants.radio import API_TYPES
from pylav.helpers.format.strings import shorten_string
from pylav.type_hints.generics import ANY_GENERIC_TYPE, PARAM_SPEC_TYPE

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLav", pathlib.Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


if TYPE_CHECKING:
    from pylav.core.client import Client, Translator
    from pylav.extension.radio.objects import Codec, Country, CountryCode, Language, State, Station, Tag


CACHE = Cache("TRANSFORMER_CACHE")
CACHE.setup("mem://?check_interval=10", size=1_000_000, enable=True)


class Error(Exception):
    """Base class for all exceptions raised by this module"""


class IllegalArgumentError(Error):
    """Raised for illegal argument"""


def bool_to_string(boolean: bool) -> str:
    """Convert a boolean type to string.

    Args:
        boolean (bool): A Boolean.

    Raises:
        TypeError

    Returns:
        str: String representation of a bool type.
    """
    string = str(boolean).lower()
    if string in {"true", "false"}:
        return string
    raise TypeError("Value must be True or False")


def snake_to_camel(string: str) -> str:
    first, *others = string.split("_")
    return "".join([first.lower(), *map(str.title, others)])


def radio_browser_adapter(**kwargs: PARAM_SPEC_TYPE.kwargs) -> PARAM_SPEC_TYPE.kwargs:
    params = {}

    for key, value in kwargs.items():
        new_key = snake_to_camel(key)
        if isinstance(kwargs[key], bool):
            value = bool_to_string(value)
        params[new_key] = value
    return params


def validate_input(
    type_value: dict[str, type[str | bool | int]] | dict[str, str | int | bool],
    input_data: PARAM_SPEC_TYPE.kwargs,
) -> None:
    for key, value in input_data.items():
        try:
            key_type = type_value[key]
        except KeyError as exc:
            raise IllegalArgumentError(f"There is no parameter named '{exc.args[0]}'") from exc
        else:
            if not isinstance(value, key_type):
                raise TypeError(f"Argument {key!r} must be {key_type.__name__}, not {type(value).__name__}")


def type_check(
    func: Callable[PARAM_SPEC_TYPE, Awaitable[ANY_GENERIC_TYPE]]  # type: ignore
) -> Callable[PARAM_SPEC_TYPE, Awaitable[ANY_GENERIC_TYPE]]:  # type: ignore
    @wraps(func)
    def wrapper(self, *args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs) -> Awaitable[ANY_GENERIC_TYPE]:
        validate_input(API_TYPES[func.__name__], kwargs)
        kwargs = radio_browser_adapter(**kwargs)
        return func(self, *args, **kwargs)

    return wrapper


class TransformerCache:
    _cache_stations: dict[str, Station] = {}
    _cache_tags: dict[str, Tag] = {}
    _cache_languages: dict[str, Language] = {}
    _cache_states: dict[str, State] = {}
    _cache_codecs: dict[str, Codec] = {}
    _cache_country_codes: dict[str, CountryCode] = {}
    _cache_countries: dict[str, Country] = {}

    _choice_cache_stations: dict[str, Choice] = {}
    _choice_cache_tags: dict[str, Choice] = {}
    _choice_cache_languages: dict[str, Choice] = {}
    _choice_cache_states: dict[str, Choice] = {}
    _choice_cache_codecs: dict[str, Choice] = {}
    _choice_cache_country_codes: dict[str, Choice] = {}
    _choice_cache_countries: dict[str, Choice] = {}

    _top_25_stations: list[Choice] = []
    _client: Client | None = None

    @classmethod
    def clear_cache(cls):
        cls._cache_stations = {}
        cls._cache_tags = {}
        cls._cache_languages = {}
        cls._cache_states = {}
        cls._cache_codecs = {}
        cls._cache_country_codes = {}
        cls._cache_countries = {}
        cls._choice_cache_stations = {}
        cls._choice_cache_tags = {}
        cls._choice_cache_languages = {}
        cls._choice_cache_states = {}
        cls._choice_cache_codecs = {}
        cls._choice_cache_country_codes = {}
        cls._choice_cache_countries = {}

    @classmethod
    async def fill_cache(cls, client: Client):
        cls._client = client
        cls._cache_stations = {
            s.stationuuid: s
            for s in sorted(
                await client.radio_browser.stations(hidebroken="true"), key=attrgetter("votes"), reverse=True
            )
        }
        cls._cache_tags = {t.name: t for t in sorted(await client.radio_browser.tags(), key=attrgetter("name"))}
        cls._cache_languages = {
            l.name: l for l in sorted(await client.radio_browser.languages(), key=attrgetter("name"))
        }
        cls._cache_states = {s.name: s for s in sorted(await client.radio_browser.states(), key=attrgetter("name"))}
        cls._cache_codecs = {c.name: c for c in sorted(await client.radio_browser.codecs(), key=attrgetter("name"))}
        cls._cache_country_codes = {
            c.name: c for c in sorted(await client.radio_browser.countrycodes(), key=attrgetter("name"))
        }
        cls._cache_countries = {
            c.name: c for c in sorted(await client.radio_browser.countries(), key=attrgetter("name"))
        }

    @classmethod
    def fill_choice_cache(cls):
        cls._choice_cache_stations = {
            station.stationuuid: Choice(
                name=shorten_string(station.name, max_length=100)
                if station.name
                else shorten_string(max_length=100, string=_("Unnamed")),
                value=f"{station.stationuuid}",
            )
            for station in cls._cache_stations.values()
        }
        cls._choice_cache_tags = {
            tag.name: Choice(
                name=shorten_string(tag.name, max_length=100)
                if tag.name
                else shorten_string(max_length=100, string=_("Unnamed")),
                value=f"{tag.name}",
            )
            for tag in cls._cache_tags.values()
        }
        cls._choice_cache_languages = {
            language.name: Choice(
                name=shorten_string(language.name, max_length=100)
                if language.name
                else shorten_string(max_length=100, string=_("Unnamed")),
                value=f"{language.name}",
            )
            for language in cls._cache_languages.values()
        }
        cls._choice_cache_states = {
            state.name: Choice(
                name=shorten_string(state.name, max_length=100) if state.name else _("Unnamed"), value=f"{state.name}"
            )
            for state in cls._cache_states.values()
        }
        cls._choice_cache_codecs = {
            codec.name: Choice(
                name=shorten_string(codec.name, max_length=100) if codec.name else _("Unnamed"), value=f"{codec.name}"
            )
            for codec in cls._cache_codecs.values()
        }
        cls._choice_cache_country_codes = {
            country_code.name: Choice(
                name=shorten_string(country_code.name, max_length=100) if country_code.name else _("Unnamed"),
                value=f"{country_code.name}",
            )
            for country_code in cls._cache_country_codes.values()
        }
        cls._choice_cache_countries = {
            country.name: Choice(
                name=shorten_string(country.name, max_length=100) if country.name else _("Unnamed"),
                value=f"{country.name}",
            )
            for country in cls._cache_countries.values()
        }

    @classmethod
    def get_station_cache(cls) -> dict[str, Station]:
        return cls._cache_stations

    @classmethod
    def get_tag_cache(cls) -> dict[str, Tag]:
        return cls._cache_tags

    @classmethod
    def get_language_cache(cls) -> dict[str, Language]:
        return cls._cache_languages

    @classmethod
    def get_state_cache(cls) -> dict[str, State]:
        return cls._cache_states

    @classmethod
    def get_codec_cache(cls) -> dict[str, Codec]:
        return cls._cache_codecs

    @classmethod
    def get_country_code_cache(cls) -> dict[str, CountryCode]:
        return cls._cache_country_codes

    @classmethod
    def get_country_cache(cls) -> dict[str, Country]:
        return cls._cache_countries

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def filter_cache(
        cls, cache_type: str, limit: int = 25, **kwargs: Any
    ) -> list[Station] | list[Tag] | list[Language] | list[State] | list[Codec] | list[CountryCode] | list[Country]:
        match cache_type:
            case "station":
                return await cls._filter_station_cache(limit=limit, **kwargs)
            case "tag":
                return await cls._filter_tag_cache(limit=limit, **kwargs)
            case "language":
                return await cls._filter_language_cache(limit=limit, **kwargs)
            case "state":
                return await cls._filter_state_cache(limit=limit, **kwargs)
            case "codec":
                return await cls._filter_codec_cache(limit=limit, **kwargs)
            case "countrycode":
                return await cls._filter_country_code_cache(limit=limit, **kwargs)
            case "country":
                return await cls._filter_country_cache(limit=limit, **kwargs)
            case _:
                return []

    @classmethod
    def build_filter(cls, **kwargs) -> dict[str, str | bool]:
        filters = {}

        if "code" in kwargs:
            filters["code"] = kwargs["code"]
        if "country" in kwargs:
            filters["country"] = kwargs["country"]
            if kwargs["country"] in cls._choice_cache_country_codes:
                filters["country_exact"] = True
                kwargs["countrycode"] = cls._cache_countries[kwargs["country"]].iso_3166_1
        if "state" in kwargs:
            filters["state"] = kwargs["state"]
            if kwargs["state"] in cls._choice_cache_states:
                filters["state_exact"] = True
        if "language" in kwargs:
            filters["language"] = kwargs["language"]
            if kwargs["language"] in cls._choice_cache_languages:
                filters["language_exact"] = True
        if "tag" in kwargs:
            filters["tag"] = kwargs["tag"]
            if kwargs["tag"] in cls._choice_cache_tags:
                filters["tag_exact"] = True
        elif "tag_list" in kwargs:
            filters["tag_list"] = kwargs["tag_list"].split(",")
        if "order" in kwargs:
            filters["order"] = kwargs["order"]
        if "countrycode" in kwargs:
            filters["countrycode"] = kwargs["countrycode"]
        if "name" in kwargs:
            filters["name"] = re.sub(r"\s+", " ", kwargs["name"])
        return filters

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def _filter_station_cache(cls, limit: int = 25, **kwargs: Any) -> list[Station]:
        filter_data = cls.build_filter(**kwargs)
        key_set = {
            "state_exact",
            "language_exact",
            "tag_exact",
            "countrycode",
            "state",
            "language",
            "tag",
            "tag_list",
            "name",
            "order",
        }

        search_args = {}
        for key in filter_data:
            if filter_data[key] is not None and key in key_set:
                data = filter_data[key]
                search_args[key] = ",".join(data) if key == "tag_list" else data
        station = await cls._client.radio_browser.search(
            limit=limit,
            **search_args,
        )

        def _sort(c: Station) -> float | list[float]:
            if "name" not in filter_data:
                return [-ord(c) for c in c.name]
            if c.name == filter_data["name"]:
                return 101
            return fuzz.partial_ratio(
                c.name,
                filter_data["name"],
            )

        return sorted(station, key=_sort, reverse=True)

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def _filter_tag_cache(cls, limit: int = 25, **kwargs: Any) -> list[Tag]:
        filter_data = cls.build_filter(**kwargs)
        tag_exact = filter_data.pop("tag_exact", False)
        tag = filter_data.pop("tag", None)
        tag_list = filter_data.pop("tag_list", None)

        def _filter_tag(c: Tag) -> float:
            if (tag_exact and c.name == tag) or (tag_list and c.name in tag_list):
                return 101
            elif tag_list:
                return 0
            return fuzz.partial_ratio(
                c.name,
                tag,
            )

        def _filter(c: Tag) -> float:
            return _filter_tag(c)

        extracted: list[Tag] = heapq.nlargest(limit, cls.get_tag_cache().values(), key=_filter)

        return extracted

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def _filter_country_cache(cls, limit: int = 25, **kwargs: Any) -> list[Country]:
        filter_data = cls.build_filter(**kwargs)
        country_exact = filter_data.pop("country_exact", False)
        country = filter_data.pop("country", None)
        countrycode = filter_data.pop("code", filter_data.pop("countrycode", None))

        def _filter_country(c: Country) -> float:
            if country_exact and c.name == country:
                return 101
            return fuzz.partial_ratio(
                c.name,
                country,
            )

        def _filter_countrycode(c: Country) -> float:
            return 101 if countrycode and c.iso_3166_1 == countrycode else 0

        def _filter(c: Country) -> float:
            result = (
                _filter_country(c) if country else 0,
                _filter_countrycode(c) if countrycode is None else 0,
            )
            return sum(result)

        extracted: list[Country] = heapq.nlargest(limit, cls.get_country_cache().values(), key=_filter)

        return extracted

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def _filter_country_code_cache(cls, limit: int = 25, **kwargs: Any) -> list[CountryCode]:
        filter_data = cls.build_filter(**kwargs)
        countrycode = filter_data.pop("code", filter_data.pop("countrycode", None))

        def _filter_countrycode(c: CountryCode) -> int:
            return 101 if countrycode and c.name == countrycode else 0

        def _filter(c: CountryCode) -> int:
            return _filter_countrycode(c) if countrycode else 0

        extracted: list[CountryCode] = heapq.nlargest(limit, cls.get_country_code_cache().values(), key=_filter)

        return extracted

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def _filter_state_cache(cls, limit: int = 25, **kwargs: Any) -> list[State]:
        filter_data = cls.build_filter(**kwargs)
        state_exact = filter_data.pop("state_exact", False)
        state = filter_data.pop("state", None)
        country = filter_data.pop("country", None)
        country_exact = filter_data.pop("country_exact", False)

        def _filter_state(c: State) -> float:
            if state_exact and c.name == state:
                return 101
            return fuzz.partial_ratio(
                c.name,
                state,
            )

        def _filter_country(c: State) -> float:
            if country_exact and c.country == country:
                return 101
            return fuzz.partial_ratio(
                c.country,
                country,
            )

        def _filter(c: State) -> float:
            result = (
                _filter_state(c) if state else 0,
                _filter_country(c) if country else 0,
            )
            return sum(result)

        extracted: list[State] = heapq.nlargest(limit, cls.get_state_cache().values(), key=_filter)

        return extracted

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def _filter_language_cache(cls, limit: int = 25, **kwargs: Any) -> list[Language]:
        filter_data = cls.build_filter(**kwargs)
        language_exact = filter_data.pop("language_exact", False)
        language = filter_data.pop("language", None)

        def _filter_language(c: Language) -> float:
            if language_exact and c.name == language:
                return 101
            return fuzz.partial_ratio(
                c.name,
                language,
            )

        def _filter(c: Language) -> float:
            return _filter_language(c) if language else 0

        extracted: list[Language] = heapq.nlargest(limit, cls.get_language_cache().values(), key=_filter)

        return extracted

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def _filter_codec_cache(cls, limit: int = 25, **kwargs: Any) -> list[Codec]:
        filter_data = cls.build_filter(**kwargs)
        codec_exact = filter_data.pop("codec_exact", False)
        codec = filter_data.pop("codec", None)

        def _filter_codec(c: Codec) -> float:
            if codec_exact and c.name == codec:
                return 101
            return fuzz.partial_ratio(
                c.name,
                codec,
            )

        def _filter(c: Codec) -> float:
            return _filter_codec(c) if codec else 0

        extracted: list[Codec] = heapq.nlargest(limit, cls.get_codec_cache().values(), key=_filter)

        return extracted

    @classmethod
    @CACHE.cache(ttl=timedelta(hours=24))
    async def get_top_25_stations(cls) -> list[Choice]:
        if cls._top_25_stations:
            return cls._top_25_stations

        def _filter(c: Station) -> int:
            return c.votes

        extracted: list[Station] = heapq.nlargest(25, cls.get_station_cache().values(), key=_filter)
        cls._top_25_stations = [
            Choice(
                name=shorten_string(e.name, max_length=100)
                if e.name
                else shorten_string(max_length=100, string=_("Unnamed")),
                value=f"{e.stationuuid}",
            )
            for e in extracted
        ]
        return cls._top_25_stations



from __future__ import annotations

import dataclasses
from typing import Literal

from pylav.nodes.api.responses.websocket import Message

__all__ = ("Segment", "SegmentsLoaded", "SegmentSkipped")


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Segment:
    category: str
    start: str
    end: str


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class SegmentsLoaded(Message):
    op: Literal["event"] = "event"
    guildId: str | None = None
    type: Literal["SegmentsLoaded"] = "SegmentsLoaded"
    segments: list[Segment | dict] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        temp = []
        for s in self.segments:
            if isinstance(s, Segment) or (isinstance(s, dict) and (s := Segment(**s))):
                temp.append(s)
        object.__setattr__(self, "segments", temp)


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class SegmentSkipped(Message):
    op: Literal["event"] = "event"
    guildId: str | None = None
    type: Literal["SegmentSkipped"] = "SegmentSkipped"
    segment: Segment | dict = dataclasses.field(default_factory=dict)

    def __post_init__(self) -> None:
        if isinstance(self.segment, dict):
            object.__setattr__(self, "segment", Segment(**self.segment))



from __future__ import annotations

from typing import TYPE_CHECKING

from packaging.version import Version

from pylav.constants.node import NODE_DEFAULT_SETTINGS
from pylav.constants.versions import VERSION_0_0_0, VERSION_0_7_6_0
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def migration_v_0_7_6_0(client: Client, current_version: Version) -> None:
    if current_version >= VERSION_0_7_6_0 or current_version == VERSION_0_0_0:
        return

    LOGGER.info("Running %s migration", VERSION_0_7_6_0)
    config = client.node_db_manager.bundled_node_config()
    yaml_data = await config.fetch_yaml()

    if len(yaml_data["lavalink"]["plugins"]) < len(NODE_DEFAULT_SETTINGS["lavalink"]["plugins"]):
        yaml_data["lavalink"]["plugins"] = NODE_DEFAULT_SETTINGS["lavalink"]["plugins"]
        await config.update_yaml(yaml_data)
    await client.lib_db_manager.update_bot_dv_version(VERSION_0_7_6_0)



from __future__ import annotations

from pathlib import Path
from typing import Any

import discord
from redbot.core.i18n import Translator

from pylav.core.context import PyLavContext
from pylav.extension.red.ui.buttons.generic import CloseButton, LabelButton, NavigateButton, RefreshButton
from pylav.extension.red.ui.buttons.player import DisconnectAllButton, DisconnectButton, StopTrackButton
from pylav.extension.red.ui.menus.generic import BaseMenu
from pylav.extension.red.ui.sources.player import PlayersSource
from pylav.helpers.format.strings import shorten_string
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class StatsMenu(BaseMenu):
    _source: PlayersSource

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: PlayersSource,
        original_author: discord.abc.User,
        *,
        clear_buttons_after: bool = False,
        delete_after_timeout: bool = True,
        timeout: int = 600,
        message: discord.Message = None,
        starting_page: int = 0,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            cog=cog,
            bot=bot,
            source=source,
            clear_buttons_after=clear_buttons_after,
            delete_after_timeout=delete_after_timeout,
            timeout=timeout,
            message=message,
            starting_page=starting_page,
            **kwargs,
        )

        self.forward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=1,
            row=0,
            cog=cog,
        )
        self.backward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=-1,
            row=0,
            cog=cog,
        )
        self.first_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}",
            direction=0,
            row=0,
            cog=cog,
        )
        self.last_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}",
            direction=self.source.get_max_pages,
            row=0,
            cog=cog,
        )
        self.refresh_button = RefreshButton(
            style=discord.ButtonStyle.grey,
            row=0,
            cog=cog,
        )

        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            row=1,
            cog=cog,
        )
        self.stop_button = StopTrackButton(
            style=discord.ButtonStyle.red,
            row=1,
            cog=cog,
        )
        self.queue_disconnect_label = LabelButton(
            disconnect_type_translation=shorten_string(max_length=100, string=_("selected")),
            row=2,
            multiple=False,
        )
        self.queue_disconnect = DisconnectButton(
            style=discord.ButtonStyle.red,
            row=2,
            cog=cog,
        )
        self.queue_disconnect_inactive_label = LabelButton(
            disconnect_type_translation=shorten_string(max_length=100, string=_("inactive")),
            row=3,
        )
        self.queue_disconnect_inactive = DisconnectAllButton(
            disconnect_type="inactive",
            style=discord.ButtonStyle.red,
            row=3,
            cog=cog,
        )
        self.queue_disconnect_all_label = LabelButton(
            disconnect_type_translation=shorten_string(max_length=100, string=_("all")),
            row=4,
        )
        self.queue_disconnect_all = DisconnectAllButton(
            disconnect_type="all",
            style=discord.ButtonStyle.red,
            row=4,
            cog=cog,
        )
        self.author = original_author

    async def prepare(self):
        self.clear_items()
        max_pages = self.source.get_max_pages()
        self.add_item(self.close_button)
        self.add_item(self.stop_button)
        self.add_item(self.queue_disconnect_label)
        self.add_item(self.queue_disconnect)
        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)
        self.add_item(self.refresh_button)
        if self.source.specified_guild is None or not (self.cog.pylav.player_manager.get(self.source.specified_guild)):
            self.add_item(self.queue_disconnect_inactive_label)
            self.add_item(self.queue_disconnect_inactive)
            self.add_item(self.queue_disconnect_all_label)
            self.add_item(self.queue_disconnect_all)

        self.forward_button.disabled = False
        self.backward_button.disabled = False
        self.first_button.disabled = False
        self.last_button.disabled = False
        self.refresh_button.disabled = False
        self.queue_disconnect.disabled = False
        self.queue_disconnect_all.disabled = False
        self.queue_disconnect_inactive.disabled = False
        self.stop_button.disabled = False
        self.queue_disconnect_label.disabled = True
        self.queue_disconnect_all_label.disabled = True
        self.queue_disconnect_inactive_label.disabled = True

        if max_pages > 2:
            self.forward_button.disabled = False
            self.backward_button.disabled = False
            self.first_button.disabled = False
            self.last_button.disabled = False
        elif max_pages == 2:
            self.forward_button.disabled = False
            self.backward_button.disabled = False
            self.first_button.disabled = True
            self.last_button.disabled = True
        elif max_pages == 1:
            self.forward_button.disabled = True
            self.backward_button.disabled = True
            self.first_button.disabled = True
            self.last_button.disabled = True
        player = self.source.current_player
        if not player:
            self.stop_button.disabled = True
            self.queue_disconnect.disabled = True
            self.queue_disconnect_label.disabled = True
            self.queue_disconnect_inactive_label.disabled = True
            self.queue_disconnect_all_label.disabled = True
        elif not player.current:
            self.stop_button.disabled = True

        if len(self.source.entries) <= 1:
            self.queue_disconnect_inactive.disabled = True
            self.queue_disconnect_all.disabled = True

        if not [p for p in iter(self.cog.pylav.player_manager.connected_players) if not p.is_active]:
            self.queue_disconnect_inactive.disabled = True

    @property
    def source(self) -> PlayersSource:
        return self._source

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(ctx, discord.Interaction):
            ctx = await self.cog.bot.get_context(ctx)
        if ctx.interaction and not ctx.interaction.response.is_done():
            await ctx.defer(ephemeral=True)
        self.ctx = ctx
        await self.send_initial_message(ctx)

    async def get_page(self, page_num: int):
        if len(self.source.entries) == 0:
            self._source.current_player = None
            return {
                "content": None,
                "embed": await self.cog.pylav.construct_embed(
                    messageable=self.ctx,
                    title=shorten_string(max_length=100, string=_("I am not in any voice channel currently.")),
                ),
            }
        try:
            if page_num >= self._source.get_max_pages():
                page_num = 0
                self.current_page = 0
            page = await self.source.get_page(page_num)
        except IndexError:
            self.current_page = 0
            page = await self.source.get_page(self.current_page)
        value = await self.source.format_page(self, page)
        if isinstance(value, dict):
            return value
        elif isinstance(value, str):
            return {"content": value, "embed": None}
        elif isinstance(value, discord.Embed):
            return {"embed": value, "content": None}



from __future__ import annotations

__ZERO__ = b"MHp2Zg=="
# noinspection SpellCheckingInspection
__MOMONGA__ = b"NTh3Yw=="
__LIGHT__ = b"|OW5hMQ=="
__SEPHIROTH__ = b"ZzRlbA=="
ANIME = b"|".join([__SEPHIROTH__, __MOMONGA__, __ZERO__, __LIGHT__])

__all__ = ("ANIME",)



import contextlib
import json
from collections.abc import Callable
from json import JSONDecodeError as JSONDecodeError
from json import JSONDecoder as JSONDecoder
from json import JSONEncoder as JSONEncoder
from types import ModuleType
from typing import IO, Any, AnyStr, overload

try:
    import orjson as _orjson

except ImportError:
    _orjson = None

try:
    import ujson as _ujson
except ImportError:
    _ujson = None


__all__ = [
    "dump",
    "dumps",
    "load",
    "loads",
    "JSONDecoder",
    "JSONDecodeError",
    "JSONEncoder",
    "get_origin",
]

# .dumps() signatures
if _orjson:

    @overload
    def dumps(
        obj: Any,
        *,
        orjson_default: Callable[[Any], Any] | None = None,
        orjson_option: int | None = None,
    ) -> str:
        ...

    __dumps_origin = _orjson

elif _ujson:

    @overload
    def dumps(
        obj: Any,
        *,
        ensure_ascii: bool = True,
        ujson_encode_html_chars: bool = False,
        ujson_escape_forward_slashes: bool = True,
        sort_keys: bool = False,
        indent: int | None = 0,
    ) -> str:
        ...

    __dumps_origin = _ujson

else:

    @overload
    def dumps(
        obj: Any,
        *,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: type[JSONEncoder] | None = None,
        indent: None | int | str = None,
        separators: tuple[str, str] | None = None,
        default: Callable[[Any], Any] | None = None,
        sort_keys: bool = False,
        **kwds: Any,
    ) -> str:
        ...

    __dumps_origin = json


# .loads() signatures
if _orjson:

    @overload
    def loads(obj: bytes | bytearray | memoryview | str) -> Any:
        ...

    __loads_origin = _orjson

elif _ujson:

    @overload
    def loads(obj: AnyStr, *, ujson_precise_float: bool | None = None) -> Any:
        ...

    __loads_origin = _ujson

else:

    @overload
    def loads(
        obj: AnyStr,
        *,
        cls: type[JSONDecoder] | None = None,
        object_hook: Callable[[dict[Any, Any]], Any] | None = None,
        parse_float: Callable[[str], Any] | None = None,
        parse_int: Callable[[str], Any] | None = None,
        parse_constant: Callable[[str], Any] | None = None,
        object_pairs_hook: Callable[[list[tuple[Any, Any]]], Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        ...

    __loads_origin = json


# .dump() signatures
if _orjson:

    @overload
    def dump(
        obj: Any,
        fp: IO[str],
    ) -> None:
        ...

    __dump_origin = _orjson

elif _ujson:

    @overload
    def dump(
        obj: Any,
        fp: IO[str],
        *,
        ensure_ascii: bool = True,
        sort_keys: bool = False,
        indent: int = 0,
        ujson_encode_html_chars: bool = False,
        ujson_escape_forward_slashes: bool = True,
    ) -> None:
        ...

    __dump_origin = _ujson

else:

    @overload
    def dump(
        obj: Any,
        fp: IO[str],
        *,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: type[JSONEncoder] | None = None,
        indent: None | int | str = None,
        separators: tuple[str, str] | None = None,
        default: Callable[[Any], Any] | None = None,
        sort_keys: bool = False,
        **kwargs: Any,
    ) -> None:
        ...

    __dump_origin = json


# .load() signatures

if _orjson:

    @overload
    def load(fp: IO[AnyStr]) -> Any:
        ...

    __load_origin = _orjson

elif _ujson:

    @overload
    def load(
        fp: IO[AnyStr],
        *,
        precise_float: bool = ...,
    ) -> Any:
        ...

    __load_origin = _ujson

else:

    @overload
    def load(
        fp: IO[AnyStr],
        *,
        cls: type[JSONDecoder] | None = ...,
        object_hook: Callable[[Any], Any] | None = ...,
        parse_float: Callable[[str], float] | None = ...,
        parse_int: Callable[[str], int] | None = ...,
        parse_constant: Callable[[str], Any] | None = ...,
        object_pairs_hook: Callable[[list[tuple[str, Any]]], Any] | None = ...,
        **kwargs: Any,
    ) -> Any:
        ...

    __load_origin = json


def dumps(
    obj: Any,
    *,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    cls=None,
    indent=None,
    separators=None,
    default=None,
    sort_keys=False,
    orjson_default=None,
    orjson_option=None,
    ujson_encode_html_chars=False,
    ujson_escape_forward_slashes=True,
    **kwargs,
) -> str:
    if _orjson:
        with contextlib.suppress(_orjson.JSONEncodeError):
            return _orjson.dumps(obj, default=orjson_default, option=orjson_option).decode()
    if _ujson:
        return _ujson.dumps(
            obj,
            ensure_ascii=ensure_ascii,
            sort_keys=sort_keys,
            indent=indent or 0,
            encode_html_chars=ujson_encode_html_chars,
            escape_forward_slashes=ujson_escape_forward_slashes,
        )
    return json.dumps(
        obj,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        cls=cls,
        indent=indent,
        separators=separators,
        default=default,
        sort_keys=sort_keys,
        **kwargs,
    )


def loads(
    obj: AnyStr | bytes | bytearray | memoryview | str,
    *,
    cls=None,
    object_hook=None,
    parse_float=None,
    parse_int=None,
    parse_constant=None,
    object_pairs_hook=None,
    ujson_precise_float=None,
    **kwargs,
) -> Any:
    if _orjson:
        with contextlib.suppress(_orjson.JSONDecodeError):
            return _orjson.loads(obj)
    if _ujson:
        with contextlib.suppress(_ujson.JSONDecodeError):
            return _ujson.loads(obj, precise_float=ujson_precise_float)
    return json.loads(
        obj,
        cls=cls,
        object_hook=object_hook,
        parse_float=parse_float,
        parse_int=parse_int,
        parse_constant=parse_constant,
        object_pairs_hook=object_pairs_hook,
        **kwargs,
    )


def dump(
    obj: Any,
    fp: IO[AnyStr],
    *,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    cls=None,
    indent=None,
    separators=None,
    default=None,
    sort_keys=False,
    ujson_encode_html_chars=False,
    ujson_escape_forward_slashes=True,
    **kwargs,
) -> None:
    if _orjson:
        with contextlib.suppress(_orjson.JSONEncodeError):
            fp.write(_orjson.dumps(obj).decode())
            return
    if _ujson:
        return _ujson.dump(
            obj,
            fp,
            ensure_ascii=ensure_ascii,
            sort_keys=sort_keys,
            indent=indent or 0,
            encode_html_chars=ujson_encode_html_chars,
            escape_forward_slashes=ujson_escape_forward_slashes,
        )
    return json.dump(
        obj,
        fp,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        cls=cls,
        indent=indent,
        separators=separators,
        default=default,
        sort_keys=sort_keys,
        **kwargs,
    )


def load(
    fp: IO[AnyStr],
    *,
    cls=None,
    object_hook=None,
    parse_float=None,
    parse_int=None,
    parse_constant=None,
    object_pairs_hook=None,
    precise_float=None,
    **kwargs,
) -> Any:
    if _orjson:
        with contextlib.suppress(_orjson.JSONDecodeError):
            return _orjson.loads(fp.read())
    if _ujson:
        with contextlib.suppress(_ujson.JSONDecodeError):
            return _ujson.load(fp, precise_float=precise_float)
    return json.load(
        fp,
        cls=cls,
        object_hook=object_hook,
        parse_float=parse_float,
        parse_int=parse_int,
        parse_constant=parse_constant,
        object_pairs_hook=object_pairs_hook,
        **kwargs,
    )


def get_origin() -> dict[str, ModuleType]:
    return {
        "dumps": __dumps_origin,
        "loads": __loads_origin,
        "dump": __dump_origin,
        "load": __load_origin,
    }



from __future__ import annotations

import random
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

from discord.app_commands import Choice, Transformer
from discord.ext import commands

from pylav.exceptions.database import EntryNotFoundException
from pylav.extension.radio.objects import Codec, Country, CountryCode, Language, State, Station, Tag
from pylav.extension.radio.utils import TransformerCache
from pylav.helpers.format.strings import shorten_string
from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE

try:
    from redbot.core.i18n import Translator  # type: ignore

    _ = Translator("PyLav", Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


if TYPE_CHECKING:
    from discord.types.interactions import ChatInputApplicationCommandInteractionData

LOGGER = getLogger("PyLav.extension.Shared.converters.radio")

if TYPE_CHECKING:
    StationConverter = TypeVar("StationConverter", bound=list[Station])
    TagConverter = TypeVar("TagConverter", bound=list[Tag])
    LanguageConverter = TypeVar("LanguageConverter", bound=list[Language])
    StateConverter = TypeVar("StateConverter", bound=list[State])
    CodecConverter = TypeVar("CodecConverter", bound=list[Codec])
    CountryCodeConverter = TypeVar("CountryCodeConverter", bound=list[CountryCode])
    CountryConverter = TypeVar("CountryConverter", bound=list[Country])
else:

    class StationConverter(Transformer, TransformerCache):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Station]:
            """Converts a station name to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                if arg in cls._cache_stations:
                    return [cls._cache_stations[arg]]
                return await cls.filter_cache(cache_type="station", limit=25, name=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _("A station with the name `{user_input_variable_do_not_translate}` was not found.").format(
                        user_input_variable_do_not_translate=arg
                    )
                ) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Station]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(
                            max_length=100, string=_("The Radio Browser extension is currently disabled.")
                        ),
                        value="???",
                    )
                ]
            data = interaction.data
            kwargs = cls.process_kwargs(current, data)
            if not current and not kwargs:
                return await cls.get_top_25_stations()
            kwargs["order"] = "votes"
            stations = await cls.filter_cache(cache_type="station", limit=25, **kwargs)
            for station in stations:
                cls.maybe_add_station_to_cache(station)
            return [cls._choice_cache_stations[stations.stationuuid] for stations in stations]

        @classmethod
        def maybe_add_station_to_cache(cls, station: Station) -> None:
            if station.stationuuid not in cls._choice_cache_stations:
                cls._choice_cache_stations[station.stationuuid] = Choice(
                    name=shorten_string(station.name, max_length=100)
                    if station.name
                    else shorten_string(max_length=100, string=_("Unnamed")),
                    value=f"{station.stationuuid}",
                )
                cls._cache_stations[station.stationuuid] = station

        @staticmethod
        def process_kwargs(
            current: str,
            data: ChatInputApplicationCommandInteractionData | None,
        ) -> dict[str, Any]:
            kwargs = {}
            if options := data.get("options", []):
                country_code = [v for v in options if v.get("name") == "countrycode"]
                country = [v for v in options if v.get("name") == "country"]
                state = [v for v in options if v.get("name") == "state"]
                language = [v for v in options if v.get("name") == "language"]
                tags = [v for v in options if v.get("name", "").startswith("tag")]
                if country_code and (val := country_code[0].get("value")):
                    kwargs["countrycode"] = val
                if country and (val := country[0].get("value")):
                    kwargs["country"] = val
                if state and (val := state[0].get("value")):
                    kwargs["state"] = val
                if language and (val := language[0].get("value")):
                    kwargs["language"] = val
                if tags:
                    if len(tags) == 1 and (val := tags[0].get("value")):
                        kwargs["tag"] = val
                    elif len(tags) > 1:
                        kwargs["tag_list"] = ",".join([tv for t in tags if (tv := t.get("value"))])
            if current:
                kwargs["name"] = current
            return kwargs

    class TagConverter(Transformer, TransformerCache):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Tag]:
            """Converts a Tag name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                if arg in cls._cache_tags:
                    return [cls._cache_tags[arg]]
                return await cls.filter_cache(cache_type="tag", limit=25, tag=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _("A tag with the name `{user_input_variable_do_not_translate}` was not found.").format(
                        user_input_variable_do_not_translate=arg
                    )
                ) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Tag]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(
                            max_length=100, string=_("The Radio Browser extension is currently disabled.")
                        ),
                        value="???",
                    )
                ]

            if not current:
                return random.choices(list(cls._choice_cache_tags.values()), k=25)

            tags = await cls.filter_cache(cache_type="tag", limit=25, tag=current)

            return [cls._choice_cache_tags[tag.name] for tag in tags]

    class LanguageConverter(Transformer, TransformerCache):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Language]:
            """Converts a Language name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                if arg in cls._cache_languages:
                    return [cls._cache_languages[arg]]
                return await cls.filter_cache(cache_type="language", limit=25, language=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _("A language with the name `{user_input_variable_do_not_translate}` was not found.").format(
                        user_input_variable_do_not_translate=arg
                    )
                ) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Language]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(
                            max_length=100, string=_("The Radio Browser extension is currently disabled.")
                        ),
                        value="???",
                    )
                ]
            if not current:
                return random.choices(list(cls._choice_cache_languages.values()), k=25)

            languages = await cls.filter_cache(cache_type="language", limit=25, language=current)

            return [cls._choice_cache_languages[language.name] for language in languages]

    class StateConverter(Transformer, TransformerCache):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[State]:
            """Converts a State name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []
            try:
                if arg in cls._cache_states:
                    return [cls._cache_states[arg]]
                return await cls.filter_cache(cache_type="state", limit=25, state=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _("A state with the name `{user_input_variable_do_not_translate}` was not found.").format(
                        user_input_variable_do_not_translate=arg
                    )
                ) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[State]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(
                            max_length=100, string=_("The Radio Browser extension is currently disabled.")
                        ),
                        value="???",
                    )
                ]
            if not current:
                return random.choices(list(cls._choice_cache_states.values()), k=25)
            states = await cls.filter_cache(cache_type="state", limit=25, state=current)
            return [cls._choice_cache_states[state.name] for state in states]

    class CodecConverter(Transformer, TransformerCache):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Codec]:
            """Converts a Codec name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                if arg in cls._cache_codecs:
                    return [cls._cache_codecs[arg]]
                return await cls.filter_cache(cache_type="codec", limit=25, codec=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _("A codec with the name `{user_input_variable_do_not_translate}` was not found.").format(
                        user_input_variable_do_not_translate=arg
                    )
                ) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Codec]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(
                            max_length=100, string=_("The Radio Browser extension is currently disabled.")
                        ),
                        value="???",
                    )
                ]
            if not current:
                return random.choices(list(cls._choice_cache_codecs.values()), k=25)
            codecs = await cls.filter_cache(cache_type="codec", limit=25, codec=current)
            return [cls._choice_cache_codecs[codec.name] for codec in codecs]

    class CountryCodeConverter(Transformer, TransformerCache):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[CountryCode]:
            """Converts a CountryCode name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                if arg in cls._cache_country_codes:
                    return [cls._cache_country_codes[arg]]
                return await cls.filter_cache(cache_type="countrycode", limit=25, countrycode=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _("A country code with the value `{user_input_variable_do_not_translate}` was not found.").format(
                        user_input_variable_do_not_translate=arg
                    )
                ) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[CountryCode]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(
                            max_length=100, string=_("The Radio Browser extension is currently disabled.")
                        ),
                        value="???",
                    )
                ]
            if not current:
                return random.choices(list(cls._choice_cache_country_codes.values()), k=25)
            country_codes = await cls.filter_cache(cache_type="countrycode", limit=25, countrycode=current)
            return [cls._choice_cache_country_codes[country_code.name] for country_code in country_codes]

    class CountryConverter(Transformer, TransformerCache):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Country]:
            """Converts a Country name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                if arg in cls._cache_countries:
                    return [cls._cache_countries[arg]]
                return await cls.filter_cache(cache_type="country", limit=25, country=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _("A country with the name `{user_input_variable_do_not_translate}` was not found.").format(
                        user_input_variable_do_not_translate=arg
                    )
                ) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Country]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(
                            max_length=100, string=_("The Radio Browser extension is currently disabled.")
                        ),
                        value="???",
                    )
                ]
            data = interaction.data
            options = data.get("options", [])
            kwargs = {}
            if options:
                code = [v for v in options if v.get("name") in ["countrycode", "code"]]
                if code and (val := code[0].get("value")):
                    kwargs["countrycode"] = val

            countries = await cls.filter_cache(cache_type="country", limit=25, country=current, **kwargs)
            return [cls._choice_cache_countries[country.name] for country in countries]



from __future__ import annotations

import random
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime

from dacite import from_dict

from pylav.compat import json
from pylav.constants.config import READ_CACHING_ENABLED
from pylav.helpers.singleton import SingletonCachedByKey
from pylav.nodes.api.responses.track import Track
from pylav.storage.database.cache.decodators import maybe_cached
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.queries import QueryRow
from pylav.storage.database.tables.tracks import TrackRow
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class Query(CachedModel, metaclass=SingletonCachedByKey):
    id: str

    def get_cache_key(self) -> str:
        return self.id

    @maybe_cached
    async def exists(self) -> bool:
        """Check if the config exists.

        Returns
        -------
        bool
            Whether the config exists.
        """
        return await QueryRow.exists().where(QueryRow.identifier == self.id)

    async def delete(self) -> None:
        """Delete the query from the database"""
        await QueryRow.delete().where(QueryRow.identifier == self.id)
        await self.invalidate_cache()

    @maybe_cached
    async def size(self) -> int:
        """Count the tracks of the playlist.

        Returns
        -------
        int
            The number of tracks in the playlist.
        """
        tracks = await self.fetch_tracks()
        return len(tracks) if tracks else 0

    @maybe_cached
    async def fetch_tracks(self) -> list[str | JSON_DICT_TYPE]:
        """Get the tracks of the playlist.

        Returns
        -------
        list[str]
            The tracks of the playlist.
        """
        response = (
            await QueryRow.select(QueryRow.tracks(TrackRow.encoded, TrackRow.info, TrackRow.pluginInfo, load_json=True))
            .where(QueryRow.identifier == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        data = response["tracks"] if response else []
        # TODO: Remove me release 1.9
        if data and any(t["info"] is None for t in data):
            tracks = await self.client.decode_tracks([t["encoded"] for t in data], raise_on_failure=True)
            await self.update_tracks(tracks)
            data = [t.to_dict() for t in tracks]
        return data

    async def update_tracks(self, tracks: list[str | Track]):
        """Update the tracks of the playlist.

        Parameters
        ----------
        tracks: list[str | Track]
            The tracks of the playlist.
        """
        query_row = await QueryRow.objects().get_or_create(QueryRow.identifier == self.id)
        try:
            old_tracks = await query_row.get_m2m(QueryRow.tracks)
        except ValueError:
            old_tracks = []
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        _temp = defaultdict(list)
        for x in tracks:
            _temp[type(x)].append(x)
        for entry_type, entry_list in _temp.items():
            if entry_type == str:
                for track_object in await self.client.decode_tracks(entry_list, raise_on_failure=False):
                    new_tracks.append(await TrackRow.get_or_create(track_object))
            elif entry_type == dict:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(from_dict(data_class=Track, data=track_object)))
            else:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(track_object))

        if old_tracks:
            await query_row.remove_m2m(*old_tracks, m2m=QueryRow.tracks)
        if new_tracks:
            await query_row.add_m2m(*new_tracks, m2m=QueryRow.tracks)
        await self.invalidate_cache(self.fetch_tracks, self.fetch_first)
        await self.update_cache(
            (self.size, len(tracks)),
            (self.exists, True),
        )

    @maybe_cached
    async def fetch_plugin_info(self) -> JSON_DICT_TYPE:
        """Get the plugin info of the playlist.

        Returns
        -------
        JSON_DICT_TYPE
            The plugin info of the playlist.
        """
        response = (
            await QueryRow.select(QueryRow.pluginInfo)
            .where(QueryRow.identifier == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return response["pluginInfo"] if response else {}

    async def update_plugin_info(self, plugin_info: JSON_DICT_TYPE) -> None:
        """Update the plugin info of the playlist.

        Parameters
        ----------
        plugin_info: JSON_DICT_TYPE
            The plugin info of the playlist.
        """

        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await QueryRow.raw(
            """
            INSERT INTO node (id, extras) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET "pluginInfo" = excluded."pluginInfo";
            """,
            self.id,
            json.dumps(plugin_info),
        )
        await self.update_cache((self.fetch_plugin_info, plugin_info), (self.exists, True))

    @maybe_cached
    async def fetch_name(self) -> str:
        """Get the name of the playlist.

        Returns
        -------
        str
            The name of the playlist.
        """
        response = (
            await QueryRow.select(QueryRow.name)
            .where(QueryRow.identifier == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return response["name"] if response else QueryRow.name.default

    async def update_name(self, name: str) -> None:
        """Update the name of the playlist.

        Parameters
        ----------
        name: str
            The name of the playlist.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await QueryRow.raw(
            "INSERT INTO query (identifier, name) VALUES ({}, {}) ON CONFLICT (identifier) DO UPDATE SET name = "
            "EXCLUDED.name;",
            self.id,
            name,
        )
        await self.update_cache((self.fetch_name, name), (self.exists, True))

    @maybe_cached
    async def fetch_last_updated(self) -> datetime:
        """Get the last updated time of the playlist.

        Returns
        -------
        datetime
            The last updated time of the playlist.
        """
        response = (
            await QueryRow.select(QueryRow.last_updated)
            .where(QueryRow.identifier == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return response["last_updated"] if response else QueryRow.last_updated.default

    async def update_last_updated(self) -> None:
        """Update the last updated time of the playlist"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await QueryRow.raw(
            "INSERT INTO query (identifier, last_updated) "
            "VALUES ({}, {}) ON CONFLICT (identifier) "
            "DO UPDATE SET last_updated = EXCLUDED.last_updated;",
            self.id,
            QueryRow.last_updated.default.python(),
        )
        await self.update_cache(
            (self.fetch_last_updated, QueryRow.last_updated.default.python()),
            (self.exists, True),
        )

    async def bulk_update(self, tracks: list[str | Track], name: str) -> None:
        """Bulk update the query.

        Parameters
        ----------
        tracks: list[str | Track]
            The tracks of the playlist.
        name: str
            The name of the playlist
        """
        defaults = {QueryRow.name: name}
        query_row = await QueryRow.objects().get_or_create(QueryRow.identifier == self.id, defaults)
        # noinspection PyProtectedMember
        if not query_row._was_created:
            await QueryRow.update(defaults).where(QueryRow.identifier == self.id)
        try:
            old_tracks = await query_row.get_m2m(QueryRow.tracks)
        except ValueError:
            old_tracks = []
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        _temp = defaultdict(list)
        for x in tracks:
            _temp[type(x)].append(x)
        for entry_type, entry_list in _temp.items():
            if entry_type == str:
                for track_object in await self.client.decode_tracks(entry_list, raise_on_failure=False):
                    new_tracks.append(await TrackRow.get_or_create(track_object))
            elif entry_type == dict:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(from_dict(data_class=Track, data=track_object)))
            else:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(track_object))
        if old_tracks:
            await query_row.remove_m2m(*old_tracks, m2m=QueryRow.tracks)
        if new_tracks:
            await query_row.add_m2m(*new_tracks, m2m=QueryRow.tracks)
        await self.invalidate_cache(self.fetch_tracks, self.fetch_first)
        await self.update_cache(
            (self.size, len(tracks)),
            (self.fetch_name, name),
            (self.fetch_last_updated, QueryRow.last_updated.default.python()),
            (self.exists, True),
        )

    async def fetch_index(self, index: int) -> JSON_DICT_TYPE | None:
        """Get the track at the index.

        Parameters
        ----------
        index: int
            The index of the track

        Returns
        -------
        str
            The track at the index
        """
        if READ_CACHING_ENABLED:
            tracks = await self.fetch_tracks()
            return tracks[index] if index < len(tracks) else None
        else:
            tracks = await self.fetch_tracks()
            if tracks and len(tracks) > index:
                return tracks[index]

    @maybe_cached
    async def fetch_first(self) -> JSON_DICT_TYPE | None:
        """Get the first track.

        Returns
        -------
        str
            The first track
        """
        return await self.fetch_index(0)

    async def fetch_random(self) -> JSON_DICT_TYPE | None:
        """Get a random track.

        Returns
        -------
        str
            A random track
        """
        return await self.fetch_index(random.randint(0, await self.size()))



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from discord.app_commands import Choice, Transformer
from discord.ext import commands
from redbot.core.i18n import Translator

from pylav.core.context import PyLavContext
from pylav.exceptions.database import EntryNotFoundException
from pylav.type_hints.bot import DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))

if TYPE_CHECKING:
    BassBoostConverter = str
else:

    class BassBoostConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: PyLavContext, arg: str) -> str:  # noqa
            """Converts user input to a valid argument for the bassboost command"""

            try:
                if match := next(
                    filter(
                        lambda x: x.lower().startswith(arg.lower()),
                        [
                            "Maximum",
                            "Insane",
                            "Extreme",
                            "High",
                            "Very High",
                            "Medium",
                            "Fine Tuned",
                            "Cut-off",
                            "Off",
                        ],
                    ),
                    None,
                ):
                    return match
                else:
                    raise EntryNotFoundException
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _(
                        "A bass boost profile with the name `{user_input_variable_do_not_translate}` was not found."
                    ).format(user_input_variable_do_not_translate=arg)
                ) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> str:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            return [
                Choice(name=t, value=p)
                for p, t in [
                    ("Maximum", _("Maximum")),
                    ("Insane", _("Insane")),
                    ("Extreme", _("Extreme")),
                    ("Very High", _("Very High")),
                    ("High", _("High")),
                    ("Medium", _("Medium")),
                    ("Fine Tuned", _("Fine Tuned")),
                    ("Cut-off", _("Cut-off")),
                    ("Off", _("Turn off")),
                ]
                if current.lower() in p.lower()
            ]



from __future__ import annotations

import os
import pathlib


def remove_keys(*keys, data: dict) -> dict:
    for key in keys:
        data.pop(key, None)
    return data


def in_container() -> bool:
    try:
        proc_1 = r"/proc/1/sched"

        out = pathlib.Path(proc_1).read_text() if os.path.exists(proc_1) else ""
        checks = [
            "docker" in out,
            "/lxc/" in out,
            out.split(" ")[0]
            not in (
                "systemd",
                "init",
            ),
            os.path.exists("./dockerenv"),
            os.path.exists("/.dockerinit"),
            os.getenv("container") is not None,
        ]
        return any(checks)
    except Exception:
        return False



from __future__ import annotations

import functools
from collections.abc import Awaitable, Callable

from pylav.constants.config import READ_CACHING_ENABLED
from pylav.storage.database.cache.cache import CACHE
from pylav.storage.database.cache.functions import key_builder
from pylav.type_hints.generics import ANY_GENERIC_TYPE, PARAM_SPEC_TYPE


def maybe_cached(
    func: Callable[PARAM_SPEC_TYPE, Awaitable[ANY_GENERIC_TYPE]]  # type: ignore
) -> Callable[ANY_GENERIC_TYPE, Awaitable[ANY_GENERIC_TYPE]]:  # type: ignore
    @functools.wraps(func)
    async def wrapper(*args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs) -> Awaitable[ANY_GENERIC_TYPE]:
        if READ_CACHING_ENABLED:
            return await CACHE(ttl=None, key=key_builder(func, *args, **kwargs))(func)(*args, **kwargs)
        return await func(*args, **kwargs)

    return wrapper



from __future__ import annotations

import asyncio
import contextlib
import datetime
import pathlib
from collections.abc import Iterable
from copy import copy
from typing import TYPE_CHECKING, Any

import discord
from discord import Attachment, Message, MessageType, PartialMessageable
from discord.ext import commands as dpy_command
from discord.ext.commands import Context as DpyContext
from discord.ext.commands.view import StringView
from discord.types.embed import EmbedType
from discord.utils import MISSING as D_MISSING  # noqa

from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_COG_TYPE, DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE
from pylav.utils.vendor.redbot import MessagePredicate

try:
    from redbot.core.commands import Command
    from redbot.core.commands import Context as OriginalContextClass
except ImportError:
    from discord.ext.commands import Command
    from discord.ext.commands import Context as OriginalContextClass


try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLav", pathlib.Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


if TYPE_CHECKING:
    from pylav.players.player import Player


class PyLavContext(OriginalContextClass):
    _original_ctx_or_interaction: DISCORD_CONTEXT_TYPE | DISCORD_INTERACTION_TYPE | None
    bot: DISCORD_BOT_TYPE
    client: DISCORD_BOT_TYPE
    interaction: DISCORD_INTERACTION_TYPE | None

    def __init__(
        self,
        *,
        message: Message,
        bot: DISCORD_BOT_TYPE,
        view: StringView,
        args: list[Any] = D_MISSING,
        kwargs: dict[str, Any] = D_MISSING,
        prefix: str | None = None,
        command: Command[Any, ..., Any] | None = None,
        invoked_with: str | None = None,
        invoked_parents: list[str] = D_MISSING,
        invoked_subcommand: Command[Any, ..., Any] | None = None,  # noqa
        subcommand_passed: str | None = None,
        command_failed: bool = False,
        current_parameter: discord.ext.commands.Parameter | None = None,
        current_argument: str | None = None,
        interaction: DISCORD_INTERACTION_TYPE | None = None,
    ):
        super().__init__(
            message=message,
            bot=bot,
            view=view,
            args=args,
            kwargs=kwargs,
            prefix=prefix,
            command=command,
            invoked_with=invoked_with,
            invoked_parents=invoked_parents,
            invoked_subcommand=invoked_subcommand,
            subcommand_passed=subcommand_passed,
            command_failed=command_failed,
            current_parameter=current_parameter,
            current_argument=current_argument,
            interaction=interaction,
        )

        self._original_ctx_or_interaction = None
        self.lavalink = bot.pylav
        self.pylav = bot.pylav

    @discord.utils.cached_property
    def author(self) -> discord.User | discord.Member:
        """Union[:class:`~discord.User`, :class:`.Member`]:
        Returns the author associated with this context's command. Shorthand for :attr:`.Message.author`
        """
        # When using client.get_context() on a button interaction the "author" becomes the bot user
        #   This ensures the original author remains the author of the context
        if isinstance(self._original_ctx_or_interaction, discord.Interaction):
            return self._original_ctx_or_interaction.user
        elif isinstance(self._original_ctx_or_interaction, DpyContext):
            return self._original_ctx_or_interaction.author
        else:
            return self.message.author

    @property
    def cog(self) -> DISCORD_COG_TYPE | None:
        """Optional[:class:`.Cog`]: Returns the cog associated with this context's command. None if it does not
        exist"""

        return None if self.command is None else self.command.cog

    @discord.utils.cached_property
    def guild(self) -> discord.Guild | None:
        """Optional[:class:`.Guild`]: Returns the guild associated with this context's command. None if not
        available"""
        return getattr(self.author, "guild", None)

    @discord.utils.cached_property
    def channel(self) -> discord.abc.MessageableChannel:
        """Union[:class:`.abc.Messageable`]: Returns the channel associated with this context's command.
        Shorthand for :attr:`.Message.channel`.
        """
        if isinstance(self._original_ctx_or_interaction, (discord.Interaction, DpyContext)):
            return self._original_ctx_or_interaction.channel  # type: ignore
        else:
            return self.message.channel

    @property
    def player(self) -> Player | None:
        """
        Get player
        """
        return self.pylav.get_player(self.guild)

    async def connect_player(
        self, channel: discord.channel.VocalGuildChannel = None, self_deaf: bool | None = None
    ) -> Player:
        """
        Connect player
        """
        requester = self.author
        channel = channel or self.author.voice.channel
        return await self.pylav.connect_player(requester=requester, channel=channel, self_deaf=self_deaf)

    @property
    def original_ctx_or_interaction(self) -> DISCORD_CONTEXT_TYPE | DISCORD_INTERACTION_TYPE | None:
        """
        Get original ctx or interaction
        """
        return self._original_ctx_or_interaction

    async def construct_embed(
        self,
        *,
        embed: discord.Embed = None,
        colour: discord.Colour | int | None = None,
        color: discord.Colour | int | None = None,
        title: str = None,
        embed_type: EmbedType = "rich",
        url: str = None,
        description: str = None,
        timestamp: datetime.datetime = None,
        author_name: str = None,
        author_url: str = None,
        thumbnail: str = None,
        footer: str = None,
        footer_url: str = None,
        messageable: discord.abc.Messageable | DISCORD_INTERACTION_TYPE = None,
    ) -> discord.Embed:
        """
        Construct embed
        """
        return await self.pylav.construct_embed(
            embed=embed,
            colour=colour,
            color=color,
            title=title,
            embed_type=embed_type,
            url=url,
            description=description,
            timestamp=timestamp,
            author_name=author_name,
            author_url=author_url,
            thumbnail=thumbnail,
            footer=footer,
            footer_url=footer_url,
            messageable=messageable or self,
        )

    @classmethod
    async def from_interaction(cls, interaction: DISCORD_INTERACTION_TYPE, /) -> PyLavContext:
        #  When using this on a button interaction it raises an error as expected.
        #   This makes the `get_context` method work with buttons by storing the original context

        added_dummy = False
        if isinstance(interaction, discord.Interaction) and interaction.command is None:
            setattr(interaction, "_cs_command", _dummy_command)
            added_dummy = True

        # Circular import
        from discord.ext.commands.bot import BotBase

        if not isinstance(interaction.client, BotBase):
            raise TypeError("Interaction client is not derived from commands.Bot or commands.AutoShardedBot")

        command = interaction.command
        if command is None:
            raise ValueError("interaction does not have command data")

        bot: DISCORD_BOT_TYPE = interaction.client  # type: ignore
        data = interaction.data  # type: ignore
        if interaction.message is None:
            synthetic_payload = {
                "id": interaction.id,
                "reactions": [],
                "embeds": [],
                "mention_everyone": False,
                "tts": False,
                "pinned": False,
                "edited_timestamp": None,
                "type": MessageType.chat_input_command
                if data.get("type", 1) == 1
                else MessageType.context_menu_command,
                "flags": 64,
                "content": "",
                "mentions": [],
                "mention_roles": [],
                "attachments": [],
            }

            if interaction.channel_id is None:
                raise RuntimeError("interaction channel ID is null, this is probably a Discord bug")

            channel = interaction.channel or PartialMessageable(
                state=interaction._state, guild_id=interaction.guild_id, id=interaction.channel_id
            )
            message = Message(state=interaction._state, channel=channel, data=synthetic_payload)  # type: ignore
            message.author = interaction.user
            message.attachments = [a for _, a in interaction.namespace if isinstance(a, Attachment)]
        else:
            message = interaction.message

        prefix = "/" if data.get("type", 1) == 1 else "\u200b"  # Mock the prefix
        ctx = cls(
            message=message,
            bot=bot,
            view=StringView(""),
            args=[],
            kwargs={},
            prefix=prefix,
            interaction=interaction,
            invoked_with=command.name,
            command=command,  # type: ignore # this will be a hybrid command, technically
        )
        interaction._baton = ctx
        ctx.command_failed = interaction.command_failed
        if added_dummy:
            ctx.command = None
        ctx._original_ctx_or_interaction = interaction
        return ctx

    def dispatch_command(
        self, message: discord.Message, command: Command, author: discord.abc.User, args: list[str], prefix: str = None
    ) -> None:
        """
        Dispatch command
        """
        command_str = f"{prefix}{command.qualified_name} {' '.join(args)}"
        msg = copy(message)
        msg.author = author
        msg.content = command_str
        self.bot.dispatch("message", msg)

    async def send_interactive(
        self, messages: Iterable[str], box_lang: str = None, timeout: int = 15, embed: bool = False
    ) -> list[discord.Message]:
        """Send multiple messages interactively.

        The user will be prompted for whether or not they would like to view
        the next message, one at a time. They will also be notified of how
        many messages are remaining on each prompt.

        Parameters
        ----------
        messages : `iterable` of `str`
            The messages to send.
        box_lang : str
            If specified, each message will be contained within a codeblock of
            this language.
        timeout : int
            How long the user has to respond to the prompt before it times out.
            After timing out, the bot deletes its prompt message.
        embed : bool
            Whether or not to send the messages as embeds.

        """
        messages = tuple(messages)
        ret = []

        for idx, page in enumerate(messages, 1):
            if box_lang is None:
                msg = (
                    await self.send(embed=await self.pylav.construct_embed(description=page, messageable=self))
                    if embed
                    else await self.send(page)
                )
            elif embed:
                msg = await self.send(
                    embed=await self.pylav.construct_embed(description=f"```{box_lang}\n{page}\n```", messageable=self)
                )
            else:
                msg = await self.send(f"```{box_lang}\n{page}\n```")
            ret.append(msg)
            n_remaining = len(messages) - idx
            if n_remaining > 0:
                match n_remaining:
                    case 1:
                        message = _("There is still 1 message remaining. Type `more` to continue.")
                    case __:
                        message = _(
                            "There are still {remaining_variable_do_not_translate} messages remaining. Type `more` to continue."
                        ).format(remaining_variable_do_not_translate=n_remaining)

                query = await self.send(message)

                try:
                    resp = await self.bot.wait_for(
                        "message",
                        check=MessagePredicate.lower_equal_to("more", self),
                        timeout=timeout,
                    )
                except asyncio.TimeoutError:
                    with contextlib.suppress(discord.HTTPException):
                        await query.delete()
                    break
                else:
                    try:
                        await self.channel.delete_messages((query, resp))
                    except (discord.HTTPException, AttributeError):
                        # In case the bot can't delete other users' messages,
                        # or is not a bot account
                        # or channel is a DM
                        with contextlib.suppress(discord.HTTPException):
                            await query.delete()
        return ret


@dpy_command.command(name="__dummy_command", hidden=True, disabled=True)
async def _dummy_command(self, context: PyLavContext) -> None:  # noqa
    """Does nothing"""



from __future__ import annotations

import inspect
from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, Any

from pylav.storage.database.cache.functions import invalidate_cache_multi, update_cache_multi

if TYPE_CHECKING:
    from pylav.core.client import Client

__CLIENT: Client | None = None


class CachedModel:
    __CLIENT: Client | None = None

    @property
    def client(self) -> Client:
        """Get the client"""
        global __CLIENT
        return self.__CLIENT or __CLIENT

    @classmethod
    def attach_client(cls, client: Client) -> None:
        global __CLIENT
        __CLIENT = cls.__CLIENT = client

    @staticmethod
    def _predicate(member: Callable) -> bool:
        """Check if the method is a cached method"""
        if getattr(member, "__name__", "_").startswith("_"):
            return False
        return bool(hasattr(member, "cache")) if inspect.ismethod(member) else False

    def get_all_methods(self) -> Iterator[Callable]:
        """Get all methods of the class"""

        return [_callable for member, _callable in inspect.getmembers(self, predicate=self._predicate)]

    async def invalidate_cache(self, *methods: Callable) -> None:
        """Invalidate the cache for the given methods if not specify all"""
        if methods:
            await invalidate_cache_multi(methods, self)
        else:
            await invalidate_cache_multi(self.get_all_methods(), self)

    async def update_cache(self, *pairs: tuple[Callable, Any]) -> None:
        """Update the cache for the specified method"""
        await update_cache_multi(pairs, self)



from __future__ import annotations

from pylav.exceptions.base import PyLavException


class SQLException(PyLavException):
    """Base exception for errors in SQL"""


class EntryNotFoundException(SQLException):
    """Raised when an entry is not found"""



from __future__ import annotations

from pathlib import Path

import discord
from redbot.core.i18n import Translator

from pylav.helpers.format.strings import shorten_string
from pylav.players.tracks.obj import Track

_ = Translator("PyLav", Path(__file__))


class QueueTrackOption(discord.SelectOption):
    def __init__(self, name: str, description: str, value: str):
        super().__init__(
            label=shorten_string(max_length=100, string=name),
            description=shorten_string(max_length=100, string=description),
            value=value,
        )

    @classmethod
    async def from_track(cls, track: Track, index: int):
        name = await track.get_track_display_name(
            max_length=100 - (2 + len(str(index + 1))), author=False, unformatted=True
        )
        label = f"{index + 1}. {name}"
        return cls(
            name=shorten_string(max_length=100, string=label),
            description=shorten_string(max_length=100, string=await track.author()),
            value=track.id,
        )


class EffectsOption(discord.SelectOption):
    def __init__(self, label: str, description: str, value: str, index: int):
        super().__init__(
            label=shorten_string(max_length=100, string=f"{index + 1}. {label}"),
            description=shorten_string(max_length=100, string=description),
            value=value,
        )


class SearchTrackOption(discord.SelectOption):
    def __init__(self, name: str, description: str, value: str):
        super().__init__(
            label=shorten_string(max_length=100, string=name),
            description=shorten_string(max_length=100, string=description),
            value=value,
        )

    @classmethod
    async def from_track(cls, track: Track, index: int):
        name = await track.get_track_display_name(
            max_length=100 - (2 + len(str(index + 1))), author=False, unformatted=True
        )
        return cls(
            name=shorten_string(max_length=100, string=f"{index + 1}. {name}"),
            description=shorten_string(max_length=100, string=await track.author()),
            value=track.id,
        )



from __future__ import annotations

import asyncio
import functools
import threading
from collections.abc import Awaitable, Callable
from typing import Any

from pylav.type_hints.generics import ANY_GENERIC_TYPE, PARAM_SPEC_TYPE

_LOCK_SINGLETON_CLASS = threading.Lock()
_LOCK_SINGLETON_CALLABLE = threading.Lock()
_LOCKS_SINGLETON_CACHE: dict[str, threading.Lock] = {}


class SingletonClass(type):
    _instances = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> SingletonClass:
        # sourcery skip: instance-method-first-arg-name
        if cls not in cls._instances:
            cls._locked_call(*args, **kwargs)
        return cls._instances[cls]

    def _locked_call(cls, *args: Any, **kwargs: Any) -> None:
        with _LOCK_SINGLETON_CLASS:
            if cls not in cls._instances:
                cls._instances[cls] = super().__call__(*args, **kwargs)


class SingletonCallable:
    _has_run = {}
    _responses = {}

    @classmethod
    def reset(cls) -> None:
        with _LOCK_SINGLETON_CALLABLE:
            cls._has_run = {}
            cls._responses = {}

    @classmethod
    def run_once(
        cls, f: Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]  # type: ignore
    ) -> Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]:  # type: ignore
        @functools.wraps(f)
        def wrapper(*args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs) -> ANY_GENERIC_TYPE:
            with _LOCK_SINGLETON_CALLABLE:
                if not cls._has_run.get(f, False):
                    cls._has_run[f] = True
                    output = f(*args, **kwargs)
                    cls._responses[f] = output
                    return output
                else:
                    return cls._responses.get(f, None)

        cls._has_run[f] = False
        return wrapper

    @classmethod
    def run_once_async(
        cls, f: Callable[PARAM_SPEC_TYPE, Awaitable[ANY_GENERIC_TYPE]]  # type: ignore
    ) -> Callable[PARAM_SPEC_TYPE, Awaitable[ANY_GENERIC_TYPE]]:  # type: ignore
        @functools.wraps(f)
        def wrapper(*args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs) -> Awaitable[ANY_GENERIC_TYPE]:
            with _LOCK_SINGLETON_CALLABLE:
                if not cls._has_run.get(f, False):
                    cls._has_run[f.__name__] = True
                    return f(*args, **kwargs)
                else:
                    return asyncio.sleep(0)

        cls._has_run[f] = False
        return wrapper


class SingletonCachedByKey(type):
    _instances: Any = {}

    @classmethod
    def _get_key(cls, mro, **kwargs: Any) -> tuple[str, ...] | None:
        singleton_key = f'{kwargs.get("id")}'
        for base in mro:
            if base.__module__.startswith("pylav.storage.models"):
                key_name = base.__name__
                if key_name in ("PlayerState", "NodeMock", "Equalizer"):
                    return
                if key_name in ("PlayerConfig", "Config"):
                    singleton_key += f".{kwargs.get('bot')}"
                return singleton_key, key_name

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        # sourcery skip: instance-method-first-arg-name
        key = cls._get_key(cls.mro(), **kwargs)
        if key not in cls._instances:
            rps = cls._locked_call(*args, **kwargs)
            if key is None:
                return rps
        return cls._instances[key]

    def _locked_call(cls, *args: Any, **kwargs: Any) -> Any:
        if super().__class__.__name__ not in _LOCKS_SINGLETON_CACHE:
            _LOCKS_SINGLETON_CACHE[super().__class__.__name__] = threading.Lock()
        with _LOCKS_SINGLETON_CACHE[super().__class__.__name__]:
            key = cls._get_key(cls.mro(), **kwargs)
            if key not in cls._instances:
                singleton = super().__call__(*args, **kwargs)
                if key is not None:
                    cls._instances[key] = singleton
                else:
                    return singleton



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from pylav.logging import getLogger
from pylav.storage.migrations.high_level.always.process_envvar_variables import process_envvar_variables
from pylav.storage.migrations.high_level.always.set_pylav_version import set_current_version
from pylav.storage.migrations.high_level.always.set_ram_value import set_correct_ram_cap
from pylav.storage.migrations.high_level.always.update_managed_node_settings import update_managed_node_settings
from pylav.storage.migrations.high_level.one_offs.v0_0_0_2 import migration_v_0_0_0_2
from pylav.storage.migrations.high_level.one_offs.v0_3_2_0 import migration_v_0_3_2_0
from pylav.storage.migrations.high_level.one_offs.v0_3_3_0 import migration_v_0_3_3_0
from pylav.storage.migrations.high_level.one_offs.v0_3_4_0 import migration_v_0_3_4_0
from pylav.storage.migrations.high_level.one_offs.v0_3_5_0 import migration_v_0_3_5_0
from pylav.storage.migrations.high_level.one_offs.v0_3_6_0 import migration_v_0_3_6_0
from pylav.storage.migrations.high_level.one_offs.v0_7_6_0 import migration_v_0_7_6_0
from pylav.storage.migrations.high_level.one_offs.v0_8_5_0 import migration_v_0_8_5_0
from pylav.storage.migrations.high_level.one_offs.v0_8_8_0 import migration_v_0_8_8_0
from pylav.storage.migrations.high_level.one_offs.v0_9_2_0 import migration_v_0_9_2_0
from pylav.storage.migrations.high_level.one_offs.v0_10_5_0 import migration_v_0_10_5_0
from pylav.storage.migrations.high_level.one_offs.v0_11_3_0 import migration_v_0_11_3_0
from pylav.storage.migrations.high_level.one_offs.v0_11_8_0 import migration_v_0_11_8_0
from pylav.storage.migrations.high_level.one_offs.v1_0_0 import migration_v_1_0_0
from pylav.storage.migrations.high_level.one_offs.v1_0_17 import migration_v_1_1_17

if TYPE_CHECKING:
    from pylav.core.client import Client
LOGGER = getLogger("PyLav.Database.Controller.Migration")


class MigrationController:
    __slots__ = ("_client", "_tasks_depend_on_node")

    def __init__(self, client: Client):
        self._client = client
        self._tasks_depend_on_node: list[asyncio.Task[None]] = []

    async def run_updates(self) -> None:
        """Run through schema migrations"""

        current_version = await self._client.lib_db_manager.get_bot_db_version().fetch_version()
        await migration_v_0_0_0_2(self._client, current_version)
        await migration_v_0_3_2_0(self._client, current_version)
        await migration_v_0_3_3_0(self._client, current_version)
        await migration_v_0_3_4_0(self._client, current_version)
        await migration_v_0_3_5_0(self._client, current_version)
        await migration_v_0_3_6_0(self._client, current_version)
        await migration_v_0_7_6_0(self._client, current_version)
        await migration_v_0_8_5_0(self._client, current_version)
        await migration_v_0_8_8_0(self._client, current_version)
        await migration_v_0_9_2_0(self._client, current_version)
        await migration_v_0_10_5_0(self._client, current_version)
        await migration_v_0_11_3_0(self._client, current_version)
        await migration_v_0_11_8_0(self._client, current_version)
        await migration_v_1_0_0(self._client, current_version)
        await migration_v_1_1_17(self._client, current_version)
        await set_current_version(self._client)
        await set_correct_ram_cap(self._client)
        await process_envvar_variables(self._client)
        await update_managed_node_settings(self._client)

    async def run_deferred_tasks_which_depend_on_node(self) -> None:
        for coro in self._tasks_depend_on_node:
            try:
                await coro
            except Exception as e:
                LOGGER.error("Error running deferred task - %s", coro.get_name(), exc_info=e)



from __future__ import annotations

import asyncio
import contextlib
import datetime
import pathlib
import typing
from collections import namedtuple
from typing import TYPE_CHECKING

import asyncpg
import discord

from pylav.constants.config import (
    TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
    TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
    TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
)
from pylav.constants.playlists import (
    BUNDLED_DEEZER_PLAYLIST_IDS,
    BUNDLED_EXTERNAL_PLAYLISTS,
    BUNDLED_PLAYLIST_IDS,
    BUNDLED_PYLAV_PLAYLISTS,
    BUNDLED_SPOTIFY_PLAYLIST_IDS,
)
from pylav.core.context import PyLavContext
from pylav.exceptions.database import EntryNotFoundException
from pylav.helpers.time import get_now_utc
from pylav.logging import getLogger
from pylav.players.query.obj import Query
from pylav.storage.database.tables.playlists import PlaylistRow
from pylav.storage.models.playlist import Playlist
from pylav.type_hints.bot import DISCORD_BOT_TYPE
from pylav.type_hints.dict_typing import JSON_DICT_TYPE

if TYPE_CHECKING:
    from pylav.core.client import Client
LOGGER = getLogger("PyLav.Database.Controller.Playlist")


try:
    from redbot.core.i18n import Translator  # type: ignore

    _ = Translator("PyLav", pathlib.Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


class PlaylistController:
    __slots__ = ("_client",)

    def __init__(self, client: Client) -> None:
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    @staticmethod
    def get_playlist(**kwargs: typing.Any) -> Playlist:
        if identifier := kwargs.pop("identifier", kwargs.pop("id", None)):
            return Playlist(id=identifier)
        else:
            raise ValueError("Playlist identifier not provided")

    async def get_bundled_playlists(self) -> list[Playlist]:
        return [
            p for playlist in BUNDLED_PLAYLIST_IDS if (p := self.get_playlist(identifier=playlist)) and await p.exists()
        ]

    async def get_playlist_by_name(self, playlist_name: str, limit: int = None) -> list[Playlist]:
        query = (
            PlaylistRow.select(PlaylistRow.id)
            .where(PlaylistRow.name.ilike(f"%{playlist_name.lower()}%"))
            .output(load_json=True, nested=True)
        )
        if limit is not None:
            query = query.limit(limit)
        playlists = await query
        if not playlists:
            raise EntryNotFoundException(
                _("laylist with the name {playlist_name_variable_do_not_translate} was not found.").format(
                    playlist_name_variable_do_not_translate=playlist_name
                )
            )
        return [self.get_playlist(**playlist) for playlist in playlists]

    async def get_playlist_by_id(self, playlist_id: int | str) -> Playlist:
        try:
            playlist_id = int(playlist_id)
            response = await PlaylistRow.exists().where(PlaylistRow.id == playlist_id)
        except ValueError as e:
            raise EntryNotFoundException(f"Playlist with id {playlist_id} not found") from e
        if response:
            return self.get_playlist(identifier=playlist_id)
        else:
            raise EntryNotFoundException(f"Playlist with id {playlist_id} not found")

    async def get_playlist_by_name_or_id(self, playlist_name_or_id: int | str, limit: int = None) -> list[Playlist]:
        try:
            return [await self.get_playlist_by_id(playlist_name_or_id)]
        except EntryNotFoundException:
            return await self.get_playlist_by_name(playlist_name_or_id, limit=limit)

    async def get_playlists_by_author(self, author: int, return_empty: bool = True) -> list[Playlist]:
        playlists = (
            await PlaylistRow.select(PlaylistRow.id)
            .where(PlaylistRow.author == author)
            .output(load_json=True, nested=True)
        )
        if playlists or return_empty:
            return [self.get_playlist(**playlist) for playlist in playlists]
        else:
            raise EntryNotFoundException(f"Playlist with author {author} not found")

    async def get_playlists_by_scope(self, scope: int, return_empty: bool = True) -> list[Playlist]:
        playlists = (
            await PlaylistRow.select(PlaylistRow.id)
            .where(PlaylistRow.scope == scope)
            .output(load_json=True, nested=True)
        )
        if playlists or return_empty:
            return [self.get_playlist(**playlist) for playlist in playlists]
        else:
            raise EntryNotFoundException(f"Playlist with scope {scope} not found")

    async def get_all_playlists(self) -> typing.AsyncIterator[Playlist]:
        for entry in await PlaylistRow.select(PlaylistRow.id).output(load_json=True, nested=True):
            yield self.get_playlist(**entry)

    async def get_external_playlists(self, *ids: int, ignore_ids: list[int] = None) -> typing.AsyncIterator[Playlist]:
        if ignore_ids is None:
            ignore_ids = []
        base_query = PlaylistRow.select(PlaylistRow.id).output(load_json=True, nested=True)
        if ids and ignore_ids:
            query = base_query.where(
                PlaylistRow.url.is_not_null() & PlaylistRow.id.in_(ids) & PlaylistRow.id.not_in(ignore_ids)
            )
        elif ignore_ids:
            query = base_query.where(PlaylistRow.url.is_not_null() & PlaylistRow.id.not_in(ignore_ids))
        else:
            query = base_query.where(PlaylistRow.url.is_not_null() & PlaylistRow.id.is_in(ids))

        for entry in await query:
            yield self.get_playlist(**entry)

    async def create_or_update_playlist(
        self,
        identifier: int,
        scope: int,
        author: int,
        name: str,
        url: str | None = None,
        tracks: list[str | JSON_DICT_TYPE] = None,
    ) -> Playlist:
        playlist = self.get_playlist(identifier=identifier)
        await playlist.bulk_update(
            scope=scope,
            author=author,
            name=name,
            url=url,
            tracks=tracks or [],
        )
        return playlist

    async def delete_playlist(self, playlist_id: int) -> None:
        await self.get_playlist(identifier=playlist_id).delete()

    async def create_or_update_global_playlist(
        self, identifier: int, author: int, name: str, url: str | None = None, tracks: list[str] = None
    ) -> Playlist:
        return await self.create_or_update_playlist(
            identifier=identifier, scope=self._client.bot.user.id, author=author, name=name, url=url, tracks=tracks
        )

    async def create_or_update_user_playlist(
        self, identifier: int, author: int, name: str, url: str | None = None, tracks: list[str] = None
    ) -> Playlist:
        return await self.create_or_update_playlist(
            identifier=identifier, scope=author, author=author, name=name, url=url, tracks=tracks
        )

    async def create_or_update_channel_playlist(
        self,
        channel: discord.abc.MessageableChannel,
        author: int,
        name: str,
        url: str | None = None,
        tracks: list[str] = None,
    ) -> Playlist:
        return await self.create_or_update_playlist(
            identifier=channel.id, scope=channel.id, author=author, name=name, url=url, tracks=tracks
        )

    async def create_or_update_guild_playlist(
        self, guild: discord.Guild, author: int, name: str, url: str | None = None, tracks: list[str] = None
    ) -> Playlist:
        return await self.create_or_update_playlist(
            identifier=guild.id, scope=guild.id, author=author, name=name, url=url, tracks=tracks
        )

    async def create_or_update_vc_playlist(
        self,
        vc: discord.channel.VocalGuildChannel,
        author: int,
        name: str,
        url: str | None = None,
        tracks: list[str] = None,
    ) -> Playlist:
        return await self.create_or_update_playlist(
            identifier=vc.id, scope=vc.id, author=author, name=name, url=url, tracks=tracks
        )

    async def get_all_for_user(
        self,
        requester: int,
        empty: bool = False,
        *,
        vc: discord.channel.VocalGuildChannel = None,
        guild: discord.Guild = None,
        channel: discord.abc.MessageableChannel = None,
    ) -> tuple[list[Playlist], list[Playlist], list[Playlist], list[Playlist], list[Playlist]]:
        """
        Gets all playlists a user has access to in a given context.

        Globals, User specific, Guild specific, Channel specific, VC specific.

        """
        global_playlists = [
            p
            for p in await self.get_playlists_by_scope(scope=self._client.bot.user.id, return_empty=True)
            if (not empty or await p.size())
        ]
        user_playlists = [
            p
            for p in await self.get_playlists_by_scope(scope=requester, return_empty=True)
            if (not empty or await p.size())
        ]
        vc_playlists = []
        guild_playlists = []
        channel_playlists = []
        if vc is not None:
            vc_playlists = [
                p
                for p in await self.get_playlists_by_scope(scope=vc.id, return_empty=True)
                if (not empty or await p.size())
            ]
        if guild is not None:
            guild_playlists = [
                p
                for p in await self.get_playlists_by_scope(scope=guild.id, return_empty=True)
                if (not empty or await p.size())
            ]
        if channel is not None:
            channel_playlists = [
                p
                for p in await self.get_playlists_by_scope(scope=channel.id, return_empty=True)
                if (not empty or await p.size())
            ]
        return global_playlists, user_playlists, guild_playlists, channel_playlists, vc_playlists

    async def get_manageable_playlists(
        self, requester: discord.abc.User, bot: DISCORD_BOT_TYPE, *, name_or_id: str | None = None
    ) -> list[Playlist]:
        if name_or_id:
            try:
                playlists = await self.get_playlist_by_name_or_id(name_or_id)
            except EntryNotFoundException:
                playlists = []
        else:
            try:
                playlists = [p async for p in self.get_all_playlists()]
            except EntryNotFoundException:
                playlists = []
        returning_list = []
        if playlists:
            for playlist in playlists:
                if await playlist.can_manage(requester=requester, bot=bot):
                    returning_list.append(playlist)
        return returning_list

    async def update_bundled_playlists(self, *playlist_ids: int) -> None:
        # NOTICE: Update the BUNDLED_PLAYLIST_IDS constant in the constants.py file
        with contextlib.suppress(asyncio.exceptions.CancelledError, asyncpg.exceptions.CannotConnectNowError):
            await self.client.node_manager.wait_until_ready()
            # noinspection PyProtectedMember
            await self.client._maybe_wait_until_bundled_node(await self.client.managed_node_is_enabled())
            # noinspection PyProtectedMember
            old_time_stamp = await self.client._config.fetch_next_execution_update_bundled_playlists()
            id_filtered = {
                playlist_id: BUNDLED_PYLAV_PLAYLISTS[playlist_id]
                for playlist_id in playlist_ids
                if playlist_id in BUNDLED_PYLAV_PLAYLISTS
            }
            if not id_filtered:
                id_filtered = BUNDLED_PYLAV_PLAYLISTS
            for playlist_id, (name, url, source) in id_filtered.items():
                try:
                    ctx = typing.cast(
                        PyLavContext,
                        namedtuple("PyLavContext", "message author")(
                            message=discord.Object(id=playlist_id), author=discord.Object(id=self._client.bot.user.id)
                        ),
                    )
                    playlist = await Playlist.from_yaml(context=ctx, url=url, scope=self._client.bot.user.id)
                    LOGGER.info("Updating bundled playlist - %s - %s", playlist_id, f"[{source}] {name}")
                except Exception as exc:
                    LOGGER.error(
                        "Built-in playlist couldn't be parsed - %s, report this error",
                        f"[{source}] {name}",
                        exc_info=exc,
                    )
                    playlist = None
                if not playlist:
                    # noinspection PyProtectedMember
                    await self.client._config.update_next_execution_update_bundled_playlists(old_time_stamp)
                    continue
            # noinspection PyProtectedMember
            await self.client._config.update_next_execution_update_bundled_playlists(
                get_now_utc() + datetime.timedelta(days=TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS)
            )
            # noinspection PyProtectedMember
            self.client._wait_for_playlists.set()

            LOGGER.info("Finished updating bundled playlists")

    async def update_bundled_external_playlists(self, *playlist_ids: int) -> None:
        with contextlib.suppress(asyncio.exceptions.CancelledError, asyncpg.exceptions.CannotConnectNowError):
            await self.client.node_manager.wait_until_ready()
            # noinspection PyProtectedMember
            await self.client._maybe_wait_until_bundled_node(await self.client.managed_node_is_enabled())
            # noinspection PyProtectedMember
            old_time_stamp = await self.client._config.fetch_next_execution_update_bundled_external_playlists()
            # NOTICE: Update the BUNDLED_PLAYLIST_IDS constant in the constants.py file
            id_filtered = {
                playlist_id: BUNDLED_EXTERNAL_PLAYLISTS[playlist_id]
                for playlist_id in playlist_ids
                if playlist_id in BUNDLED_EXTERNAL_PLAYLISTS
            }
            if not id_filtered:
                id_filtered = BUNDLED_EXTERNAL_PLAYLISTS
            for playlist_id, (identifier, name, album_playlist) in id_filtered.items():
                # noinspection PyProtectedMember
                if (playlist_id in BUNDLED_SPOTIFY_PLAYLIST_IDS and not self.client._spotify_auth) or (
                    playlist_id in BUNDLED_DEEZER_PLAYLIST_IDS and not self.client._has_deezer_support
                ):
                    continue
                elif playlist_id in BUNDLED_SPOTIFY_PLAYLIST_IDS:
                    url = f"https://open.spotify.com/{album_playlist}/{identifier}"
                elif playlist_id in BUNDLED_DEEZER_PLAYLIST_IDS:
                    url = f"https://www.deezer.com/en/{album_playlist}/{identifier}"
                else:
                    LOGGER.debug("Unknown playlist id: %s", playlist_id)
                track_list = []
                try:
                    LOGGER.info("Updating bundled external playlist - %s - %s", playlist_id, name)
                    query = await Query.from_string(url)
                    data = await self.client.get_tracks(query, bypass_cache=True)
                    name = (
                        f"[{query.source_abbreviation}] {data.playlistInfo.name}"
                        if data.playlistInfo.name
                        else f"[{query.source_abbreviation}] {name}"
                    )
                    tracks_raw = data.tracks
                    track_list = [t_ for t in tracks_raw if (t_ := t.encoded)]
                except Exception as exc:
                    LOGGER.error(
                        "Built-in external playlist couldn't be parsed - %s, report this error", name, exc_info=exc
                    )
                    data = None
                if not data:
                    # noinspection PyProtectedMember
                    await self.client._config.update_next_execution_update_bundled_external_playlists(old_time_stamp)
                    continue
                if track_list:
                    await self.create_or_update_global_playlist(
                        identifier=playlist_id, name=name, tracks=track_list, author=self._client.bot.user.id, url=url
                    )
                else:
                    await self.delete_playlist(playlist_id=playlist_id)
            # noinspection PyProtectedMember
            await self.client._config.update_next_execution_update_bundled_external_playlists(
                get_now_utc() + datetime.timedelta(days=TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS)
            )
            LOGGER.info("Finished updating bundled external playlists")

    async def update_external_playlists(self, *playlist_ids: int) -> None:
        with contextlib.suppress(asyncio.exceptions.CancelledError, asyncpg.exceptions.CannotConnectNowError):
            await self.client.node_manager.wait_until_ready()
            # noinspection PyProtectedMember
            await self.client._maybe_wait_until_bundled_node(await self.client.managed_node_is_enabled())

            async for playlist in self.get_external_playlists(*playlist_ids, ignore_ids=BUNDLED_PLAYLIST_IDS):
                name = await playlist.fetch_name()
                url = await playlist.fetch_url()
                query = await Query.from_string(url)
                try:
                    LOGGER.info("Updating external playlist - %s (%s)", name, playlist.id)
                    response = await self.client.get_tracks(
                        query,
                        bypass_cache=True,
                    )
                    tracks_raw = response.tracks
                    track_list = [t for t in tracks_raw if t.encoded]
                    new_name = response.playlistInfo.name
                    new_name = f"[{query.source_abbreviation}] {new_name}" if new_name else None
                    if track_list:
                        await playlist.update_tracks(tracks=track_list)
                    if new_name and new_name != name:
                        await playlist.update_name(new_name)
                except Exception as exc:
                    LOGGER.error(
                        "External playlist couldn't be updated - %s (%s), report this error",
                        name,
                        playlist.id,
                        exc_info=exc,
                    )
            # noinspection PyProtectedMember
            await self.client._config.update_next_execution_update_external_playlists(
                get_now_utc() + datetime.timedelta(days=TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS)
            )
            LOGGER.info("Finished updating external playlists")

    @staticmethod
    async def count() -> int:
        """Returns the number of playlists in the database."""
        return await PlaylistRow.count()



from __future__ import annotations

from pathlib import Path

import discord
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import humanize_number
from redbot.vendored.discord.ext import menus

from pylav.extension.red.ui.menus.generic import BaseMenu
from pylav.extension.red.utils import rgetattr
from pylav.logging import getLogger
from pylav.players.player import Player
from pylav.type_hints.bot import DISCORD_COG_TYPE

LOGGER = getLogger("PyLav.ext.red.ui.sources.player")
_ = Translator("PyLav", Path(__file__))


class PlayersSource(menus.ListPageSource):
    def __init__(self, cog: DISCORD_COG_TYPE, specified_guild: int = None):
        super().__init__([], per_page=1)
        self.cog = cog
        self.current_player = None
        self.specified_guild = specified_guild

    @property
    def entries(self) -> list[Player]:
        if self.specified_guild is not None and (player := self.cog.pylav.player_manager.get(self.specified_guild)):
            return [player]
        return self.cog.pylav.player_manager.connected_players

    @entries.setter
    def entries(self, players: list[Player]):
        pass

    def get_max_pages(self):
        if self.specified_guild is not None and (player := self.cog.pylav.player_manager.get(self.specified_guild)):
            players = [player]
        else:
            players = self.cog.pylav.player_manager.connected_players
        pages, left_over = divmod(len(players), self.per_page)
        if left_over:
            pages += 1
        return pages or 1

    def get_starting_index_and_page_number(self, menu: BaseMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: BaseMenu, player: Player) -> discord.Embed:
        idx_start, page_num = self.get_starting_index_and_page_number(menu)
        connect_dur = discord.utils.format_dt(player.connected_at, style="R")
        self.current_player = player
        guild_name = player.guild.name
        queue_len = player.queue.size()
        history_queue_len = player.history.size()
        server_owner = f"{player.guild.owner} ({player.guild.owner.id})"
        current_track = (
            await player.current.get_track_display_name(max_length=50, with_url=True)
            if player.current
            else _("I am not currently playing anything on this server.")
        )

        listener_count = sum(True for m in iter(rgetattr(player, "channel.members", [])) if not m.bot)
        listeners = humanize_number(listener_count)
        current_track += "\n"

        match queue_len:
            case 1:
                queue_length_translation = _("1 track")
            case 0:
                queue_length_translation = _("0 tracks")
            case __:
                queue_length_translation = _("{queue_length_variable_do_not_translate} tracks").format(
                    queue_length_variable_do_not_translate=humanize_number(queue_len)
                )

        match history_queue_len:
            case 1:
                history_queue_length_translation = _("1 track")
            case 0:
                history_queue_length_translation = _("0 tracks")
            case __:
                history_queue_length_translation = _("{history_queue_length_variable_do_not_translate} tracks").format(
                    history_queue_length_variable_do_not_translate=humanize_number(history_queue_len)
                )

        field_values = "\n".join(
            f"**{i[0]}**: {i[1]}"
            for i in [
                (_("Server Owner"), server_owner),
                (_("Connected"), connect_dur),
                (_("Users in Voice Channel"), listeners),
                (
                    _("Queue Length"),
                    queue_length_translation,
                ),
                (
                    _("Queue History Length"),
                    history_queue_length_translation,
                ),
            ]
        )

        current_track += field_values

        embed = await self.cog.pylav.construct_embed(messageable=menu.ctx, title=guild_name, description=current_track)

        total_number_of_entries = len(self.entries)
        current_page = humanize_number(page_num + 1)
        total_number_of_pages = humanize_number(self.get_max_pages())

        match total_number_of_entries:
            case 1:
                message = _("Page 1 / 1 | 1 server")
            case 0:
                message = _("Page 1 / 1 | 0 servers")
            case __:
                message = _(
                    "Page {current_page_variable_do_not_translate} / {total_number_of_pages_variable_do_not_translate} | {total_number_of_entries_variable_do_not_translate} servers"
                ).format(
                    current_page_variable_do_not_translate=current_page,
                    total_number_of_pages_variable_do_not_translate=total_number_of_pages,
                    total_number_of_entries_variable_do_not_translate=humanize_number(total_number_of_entries),
                )

        embed.set_footer(text=message)
        return embed
